from __future__ import annotations
import json
import time
from abc import ABC, abstractmethod
from queue import Queue
from threading import Lock
from typing import Any, Callable, Type
from edsl.data import CRUDOperations, CRUD
from edsl.exceptions import LanguageModelResponseNotJSONError
from edsl.language_models.schemas import model_prices
from edsl.trackers.TrackerAPI import TrackerAPI
from edsl.utilities import repair_json


class LanguageModel(ABC):
    """ABC for LLM subclasses."""

    def __init__(self, crud: CRUDOperations = CRUD, **kwargs):
        """
        Attributes:
        - all attributes inherited from subclasses
        - lock: lock for this model to ensure TODO
        - api_queue: queue that records messages about API calls the model makes. Used by `InterviewManager` to update details about state of model.
        """
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.lock = Lock()
        self.api_queue = Queue()
        self.crud = crud

    #######################
    # CORE METHODS
    #######################
    @abstractmethod
    def execute_model_call(
        self, prompt: str, system_prompt: str = ""
    ) -> dict[str, Any]:
        """Calls the LLM's API and returns the API response."""
        raise NotImplementedError

    @abstractmethod
    def parse_response(raw_response: dict[str, Any]) -> str:
        """Parses the API response and returns the response text."""
        raise NotImplementedError

    def get_raw_response(self, prompt: str, system_prompt: str = "") -> dict[str, Any]:
        """Calls the LLM's API and returns the API response. If self.use_cache is True, then attempts to retrieve the response from the database; if not in the DB, calls the LLM and writes the response to the DB."""
        start_time = time.time()

        # called without cache
        if not self.use_cache:
            response = self.execute_model_call(prompt, system_prompt)
            with self.lock:
                response["cached_response"] = False

        # called with cache
        if self.use_cache:
            # attempt to get cached response from DB
            cached_response = self.crud.get_LLMOutputData(
                model=str(self.model),
                parameters=str(self.parameters),
                system_prompt=system_prompt,
                prompt=prompt,
            )
            # if cached response in DB, load
            if cached_response:
                response = json.loads(cached_response)
                with self.lock:
                    response["cached_response"] = True
            # otherwise, call model and save response to DB
            else:
                response = self.execute_model_call(prompt, system_prompt)
                try:
                    output = json.dumps(response)
                except json.JSONDecodeError:
                    raise LanguageModelResponseNotJSONError
                self.crud.write_LLMOutputData(
                    model=str(self.model),
                    parameters=str(self.parameters),
                    system_prompt=system_prompt,
                    prompt=prompt,
                    output=output,
                )
                with self.lock:
                    response["cached_response"] = False

        # record usage details
        end_time = time.time()
        response["elapsed_time"] = end_time - start_time
        response["timestamp"] = end_time
        self._post_tracker_event(response)

        return response

    def get_response(self, prompt: str, system_prompt: str = ""):
        """Get response, parse, and return as string."""

        def load_json_closure(json_repair_func: Callable) -> Callable:
            """
            Closure to load json, and repair if necessary.

            Notes:
            - Why this convoluted closure structure you might ask? Because it's more extensible and it captures the sequence of attemps/tranformations. We can add an LLM repair step with the best available model if the "simpler" approach of just calling repair_json doesn't work.
            """
            prior_strings = []

            def func(json_string):
                prior_strings.append(json_string)
                if len(prior_strings) > 2:
                    raise Exception(
                        f"Could not repair json.loads. Sequence of attempts: {prior_strings}"
                    )
                try:
                    dict_response = json.loads(json_string)
                    return dict_response
                except json.JSONDecodeError:
                    print("Could not load. Trying to repair.")
                    new_json = json_repair_func(json_string)
                    return func(new_json)

            return func

        raw_response = self.get_raw_response(prompt, system_prompt)
        response = self.parse_response(raw_response)
        load_json = load_json_closure(repair_json)
        return load_json(response)

    #######################
    # USEFUL METHODS
    #######################
    def _post_tracker_event(self, raw_response: dict[str, Any]) -> None:
        """Parses the API response and sends usage details to the API Queue."""
        usage = raw_response.get("usage", {})
        usage.update(
            {
                "cached_response": raw_response.get("cached_response", None),
                "elapsed_time": raw_response.get("elapsed_time", None),
                "timestamp": raw_response.get("timestamp", None),
            }
        )
        event = TrackerAPI.APICallDetails(details=usage)
        self.api_queue.put(event)

    def cost(self, raw_response: dict[str, Any]) -> float:
        """Returns the dollar cost of a raw response."""
        keys = raw_response["usage"].keys()
        prices = model_prices.get(self.model)
        return sum([prices.get(key, 0.0) * raw_response["usage"][key] for key in keys])

    #######################
    # FORWARD METHODS
    #######################
    def get_value(self, jobs) -> list[Type[LanguageModel]]:
        """Used by the Jobs class to get the value of the attribute."""
        return jobs.models

    def set_value(self, jobs, new_values: list[Type[LanguageModel]]) -> None:
        """Used by the Jobs class to set the value of the attribute."""
        jobs.models = new_values

    #######################
    # SERIALIZATION METHODS
    #######################
    def to_dict(self) -> dict[str, Any]:
        """Converts instance to a dictionary."""
        return {"model": self.model, "parameters": self.parameters}

    @classmethod
    def from_dict(cls, data: dict) -> Type[LanguageModel]:
        """Converts dictionary to a LanguageModel child instance."""
        if data["model"] == "gpt-3.5-turbo":
            from edsl.language_models import LanguageModelOpenAIThreeFiveTurbo

            model_class = LanguageModelOpenAIThreeFiveTurbo
        if data["model"] == "gpt-4":
            from edsl.language_models import LanguageModelOpenAIFour

            model_class = LanguageModelOpenAIFour
        data["use_cache"] = True
        return model_class(**data)

    #######################
    # DUNDER METHODS
    #######################
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(model = '{self.model}', parameters={self.parameters})"

    def __add__(self, other_model: Type[LanguageModel]) -> Type[LanguageModel]:
        """Combine two models into a single model (other_model takes precedence over self)"""
        print(
            f"""Warning: one model is replacing another. If you want to run both models, use a single `by` e.g., 
              by(m1, m2, m3) not by(m1).by(m2).by(m3)."""
        )
        return other_model or self



import textwrap
from pydantic import BaseModel, Field, field_validator
from simpleeval import simple_eval  # type: ignore
from typing import Optional, Type
from edsl.questions.experimental.QuestionShortAnswer import QuestionShortAnswer
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings
from edsl.utilities.utilities import random_string


class QuestionShortAnswerNumeric(QuestionShortAnswer):
    question_type: str = "short_answer_numeric"

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked the following question requiring a numeric response: 
        {{question_text}}
        {% if expression_constraint %}
        Your response must be numeric and satisfy the following constraints:
        {{expression_constraint}}
        {% endif %}
        Return a valid JSON formatted exactly like this: 
        {"answer": "<put numeric answer here>", "comment": "<put explanation here>"}  
        Example of a correctly formatted response:
        {"answer": "100", "comment": "This is my explanation..."}          
        """
        )

    def __repr__(self):
        return f"""{self.__class__.__name__}(question_text = "{self.question_text}", expression_constraints = {self.expression_constraints}, question_name = "{self.question_name}")"""

    @classmethod
    def construct_question_data_model(cls) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionShortAnswerNumeric"""

            question_text: str = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            expression_constraint: str = Field(
                ..., min_length=0, max_length=Settings.MAX_EXPRESSION_CONSTRAINT_LENGTH
            )
            allow_nonresponse: Optional[bool] = None

        return LocalQuestionData

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: str = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise ValueError("You must provide a response.")
                return value

            @field_validator("answer")
            def check_answer_numeric(cls, value):
                if not value.isnumeric():
                    raise ValueError(f"Answer should be numeric.")
                return value

            @field_validator("answer")
            def check_expression_constraint(cls, value):
                numeric_value = int(value) if value.isdigit() else float(value)
                if not simple_eval(
                    self.expression_constraint, names={"answer": numeric_value}
                ):
                    raise Exception("Answer does not match expression constraint.")
                return value

        return LocalAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        return {"answer": random_string()}

    def form_elements(self):
        raise NotImplementedError


if __name__ == "__main__":
    q = QuestionShortAnswerNumeric(
        question_text="How many astronauts do you think there are in the world right now?",
        expression_constraint="answer > 0",
        question_name="astronauts",
    )

    from edsl import Agent

    a_nasa_engineer = Agent(traits={"occupation": """NASA engineer"""})
    a_fashion_model = Agent(traits={"occupation": """fashion model"""})



import random
import textwrap
from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Type, Union
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.exceptions import QuestionAnswerValidationError
from edsl.utilities.utilities import random_string


class QuestionBudget(QuestionData):
    """Pydantic data model for QuestionBudget"""

    question_options: list[str] = Field(
        ...,
        min_length=Settings.MIN_NUM_OPTIONS,
        max_length=Settings.MAX_NUM_OPTIONS,
    )
    budget_sum: int

    def __new__(cls, *args, **kwargs) -> "QuestionBudgetEnhanced":
        # see QuestionFreeText for an explanation of how __new__ works
        instance = super(QuestionBudget, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionBudgetEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)

    @field_validator("question_options")
    def check_unique(cls, value):
        return cls.base_validator_check_unique(value)

    @field_validator("question_options")
    def check_option_string_lengths(cls, value):
        return cls.base_validator_check_option_string_lengths(value)


class QuestionBudgetEnhanced(Question):
    question_type = "budget"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted as follows, with a dictionary for your "answer"
        where the keys are the option numbers and the values are the amounts you want 
        to allocate to the options, and the sum of the values is {{budget_sum}}:
        {"answer": {<put dict of option numbers and allocation amounts here>},
        "comment": "<put explanation here>"}
        Example response for a budget of 100 and 4 options: 
        {"answer": {"0": 25, "1": 25, "2": 25, "3": 25},
        "comment": "I allocated 25 to each option."}
        There must be an allocation listed for each item (including 0).
        """
        )

    def construct_answer_data_model(self) -> Type[BaseModel]:
        budget_sum = self.budget_sum
        acceptable_answer_keys = set(range(len(self.question_options)))

        class QuestionBudgetAnswerDataModel(AnswerData):
            answer: dict[int, Union[int, float]]

            @model_validator(mode="after")
            def check_answer(self):
                current_sum = sum(self.answer.values())
                if not current_sum == budget_sum:
                    raise QuestionAnswerValidationError(
                        f"Budget sum must be {budget_sum}, but got {current_sum}."
                    )
                if any(v < 0 for v in self.answer.values()):
                    raise QuestionAnswerValidationError(
                        f"Budget values must be positive, but got {self.answer.keys()}."
                    )
                if any(
                    [key not in acceptable_answer_keys for key in self.answer.keys()]
                ):
                    raise QuestionAnswerValidationError(
                        f"Budget keys must be in {acceptable_answer_keys}, but got {self.answer.keys()}"
                    )
                if acceptable_answer_keys != set(self.answer.keys()):
                    missing_keys = acceptable_answer_keys - set(self.answer.keys())
                    raise QuestionAnswerValidationError(
                        f"All but keys must be represented in the answer. Missing: {missing_keys}"
                    )
                return self

        return QuestionBudgetAnswerDataModel

    ################
    # Less important
    ################

    def translate_answer_code_to_answer(
        self, answer_codes: dict[str, int], scenario=None
    ):
        """Translates the answer codes to the actual answers.
        For example, for a budget question with options ["a", "b", "c"],
        the answer codes are 0, 1, and 2. The LLM will respond with 0.
        This code will translate that to "a".
        """
        translated_codes = []
        for answer_code, response in answer_codes.items():
            translated_codes.append({self.question_options[int(answer_code)]: response})

        return translated_codes

    def simulate_answer(self, human_readable=True):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        if human_readable:
            keys = self.question_options
        else:
            keys = range(len(self.question_options))
        values = [random.randint(0, 100) for _ in range(len(self.question_options))]
        current_sum = sum(values)
        modified_values = [v * self.budget_sum / current_sum for v in values]
        answer = dict(zip(keys, modified_values))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    def form_elements(self):
        html_output = f"""
        <label>{self.question_text}</label>
        <p>Total Budget: {self.budget_sum}</p>\n"""
        for index, option in enumerate(self.question_options):
            html_output += f"""
            <div id="{self.question_name}_div_{index}">
                <label for="{self.question_name}_{index}">{option}</label>
                <input type="number" id="{self.question_name}_{index}" 
                    name="{self.question_name}_{index}" 
                    min="0" max="{self.budget_sum}" 
                    oninput="validateBudget(this, {self.budget_sum})">
            </div>\n"""

        html_output += f"""
        <script>
            function validateBudget(input, maxBudget) {{
                let total = 0;
                let inputs = document.querySelectorAll('input[type=number]');
                inputs.forEach(el => total += Number(el.value));
                if (total > maxBudget) {{
                    alert('Total allocation exceeds the budget!');
                    input.value = '';
                }}
            }}
        </script>
        """
        return html_output


# if __name__ == "__main__":
#     # for testing
#     from edsl.questions import QuestionBudget

#     q = QuestionBudget(
#         question_text="How would you allocate $100?",
#         question_options=["Pizza", "Ice Cream", "Burgers", "Salad"],
#         budget_sum=100,
#         question_name="food_budget",
#     )

#     q.formulate_prompt()
#     q.question_options
#     q.simulate_answer()

#     results = q.run()



import random
import textwrap
from jinja2 import Template
from pydantic import BaseModel, Field, field_validator, model_validator, root_validator
from typing import Type
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.exceptions import (
    QuestionAnswerValidationError,
    QuestionCreationValidationError,
)
from edsl.utilities.utilities import random_string


class QuestionRank(QuestionData):
    """Pydantic data model for QuestionRank"""

    question_options: list[str] = Field(
        ...,
        min_length=Settings.MIN_NUM_OPTIONS,
        max_length=Settings.MAX_NUM_OPTIONS,
    )
    # default = len(question_options) through set_default_num_selections
    num_selections: int = None

    def __new__(cls, *args, **kwargs) -> "QuestionRankEnhanced":
        # see QuestionFreeText for an explanation of how __new__ works
        instance = super(QuestionRank, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionRankEnhanced(instance)

    @root_validator(pre=True)
    def set_default_num_selections(cls, values):
        if "num_selections" not in values or values["num_selections"] is None:
            if "question_options" in values:
                values["num_selections"] = len(values["question_options"])
        return values

    @field_validator("question_options")
    def check_unique(cls, value):
        return cls.base_validator_check_unique(value)

    @field_validator("question_options")
    def check_option_string_lengths(cls, value):
        return cls.base_validator_check_option_string_lengths(value)

    @model_validator(mode="after")
    def check_options_count(self):
        if self.num_selections > len(self.question_options):
            raise QuestionCreationValidationError(
                f"Required selections = {self.num_selections}, but there are only {len(self.question_options)} options."
            )
        return self


class QuestionRankEnhanced(Question):
    question_type = "rank"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted like this, selecting the numbers of the options in order of preference, 
        with the most preferred option first, and the least preferred option last: 
        {"answer": [<put comma-separated list of answer codes here>], "comment": "<put explanation here>"}
        Exactly {{num_selections}} options must be selected.
        """
        )

    def construct_answer_data_model(self) -> Type[BaseModel]:
        acceptable_values = range(len(self.question_options))

        class QuestionRankAnswerDataModel(AnswerData):
            answer: list[int]

            @field_validator("answer")
            def check_answers_valid(cls, value):
                if all([v in acceptable_values for v in value]):
                    return value
                else:
                    raise QuestionAnswerValidationError(
                        f"Rank answer {value} has elements not in {acceptable_values}."
                    )

            @field_validator("answer")
            def check_answers_count(cls, value):
                if len(value) != self.num_selections:
                    raise QuestionAnswerValidationError(
                        f"Rank answer {value}, but exactly {self.num_selections} selections required."
                    )
                return value

        return QuestionRankAnswerDataModel

    def translate_answer_code_to_answer(self, answer_codes, scenario):
        """Translates the answer code to the actual answer."""
        if scenario is None:
            scenario = dict()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]

        translated_codes = []
        for answer_code in answer_codes:
            translated_codes.append(translated_options[int(answer_code)])
        return translated_codes

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            answer = [random.choice(self.question_options)]
        else:
            answer = [random.choice(range(len(self.question_options)))]
        return {
            "answer": answer,
            "comment": random_string(),
        }

    def form_elements(self):
        # TODO: revise this to not use dropdown? Need to collect selections
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        for index, option in enumerate(self.question_options):
            html_output += f"""<div id = "{self.question_name}_div_{index}">
            <label for="{self.question_name}_{index}">{option}</label>
            <select id="{self.question_name}_{index}" name="{self.question_name}_{index}">
            """
            for rank in range(1, self.num_selections + 1):
                html_output += f'    <option value="{rank}">{rank}</option>\n'
            html_output += "  </select>\n</div>\n"

        return html_output



import random
import textwrap
from jinja2 import Template
from pydantic import BaseModel, Field, field_validator
from typing import Optional, Type
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.exceptions import QuestionAnswerValidationError
from edsl.utilities.utilities import random_string


class QuestionMultipleChoice(QuestionData):
    """Pydantic data model for QuestionMultipleChoice"""

    question_options: list[str] = Field(
        ...,
        min_length=Settings.MIN_NUM_OPTIONS,
        max_length=Settings.MAX_NUM_OPTIONS,
    )

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionMultipleChoiceEnhanced":
        instance = super(QuestionMultipleChoice, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionMultipleChoiceEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)

    @field_validator("question_options")
    def check_unique(cls, value):
        return cls.base_validator_check_unique(value)

    @field_validator("question_options")
    def check_option_string_lengths(cls, value):
        return cls.base_validator_check_option_string_lengths(value)


class QuestionMultipleChoiceEnhanced(Question):
    question_type = "multiple_choice"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted like this, selecting only the number of the option: 
        {"answer": <put answer code here>, "comment": "<put explanation here>"}
        Only 1 option may be selected.
        """
        )

    def construct_answer_data_model(self) -> Type[BaseModel]:
        "Constructs the answer data model for this question"
        acceptable_values = range(len(self.question_options))

        class QuestionMultipleChoiceAnswerDataModel(AnswerData):
            answer: int
            comment: Optional[str] = None

            @field_validator("answer")
            def check_answer(cls, value):
                if value in acceptable_values:
                    return value
                else:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} not in acceptable values {acceptable_values}"
                    )

        return QuestionMultipleChoiceAnswerDataModel

    ################
    # Less important
    ################

    def translate_answer_code_to_answer(self, answer_code, scenario=None):
        """
        Translates the answer code to the actual answer.
        For example, for question_options ["a", "b", "c"], the answer codes are 0, 1, and 2.
        The LLM will respond with 0, and this code will translate that to "a".
        # TODO: REMOVE
        >>> q = QuestionMultipleChoice(question_text = "How are you?", question_options = ["Good", "Great", "OK", "Bad"], question_name = "how_feeling")
        >>> q.translate_answer_code_to_answer(0, {})
        'Good'
        """
        scenario = scenario or dict()
        translated_options = [
            Template(str(option)).render(scenario) for option in self.question_options
        ]
        return translated_options[int(answer_code)]

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            answer = random.choice(self.question_options)
        else:
            answer = random.choice(range(len(self.question_options)))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    def form_elements(self) -> str:
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        for index, option in enumerate(self.question_options):
            html_output += f"""
            <div id = "{self.question_name}_div_{index}">
                <input type="radio" id="{self.question_name}_{index}" name="{self.question_text}" value="{option}">
                <label for="{self.question_name}_{index}">{option}</label>
            </div>\n
            """
        return html_output


if __name__ == "__main__":
    from edsl.questions import QuestionMultipleChoice

    q = QuestionMultipleChoice(
        question_text="Do you enjoying eating custard while skydiving?",
        question_options=["yes, somtimes", "no", "only on Tuesdays"],
        question_name="goose_fight",
    )
    results = q.run()
    print(results)



from pydantic import BaseModel, Field, model_validator, field_validator
from typing import Type
from edsl.exceptions import QuestionCreationValidationError
from edsl.questions import QuestionData, Settings
from edsl.questions.QuestionCheckBox import QuestionCheckBoxEnhanced


class QuestionTopK(QuestionData):
    """Pydantic data model for QuestionTopK"""

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )
    question_options: list[str] = Field(
        ...,
        min_length=Settings.MIN_NUM_OPTIONS,
        max_length=Settings.MAX_NUM_OPTIONS,
    )
    min_selections: int
    max_selections: int

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs):
        instance = super(QuestionTopK, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionTopKEnhanced(instance)

    @field_validator("question_options")
    def check_unique(cls, value):
        return cls.base_validator_check_unique(value)

    @field_validator("question_options")
    def check_option_string_lengths(cls, value):
        return cls.base_validator_check_option_string_lengths(value)

    @model_validator(mode="after")
    def check_equal_min_max(self, value):
        if self.min_selections != self.max_selections:
            raise QuestionCreationValidationError(
                f"TopK question must have equal min and max selections"
            )
        return self

    @model_validator(mode="after")
    def check_min_within_bounds(self, value):
        if self.min_selections > len(self.question_options):
            raise QuestionCreationValidationError(
                f"TopK question must have min selections less than or equal to the number of options"
            )
        if self.min_selections <= 0:
            raise QuestionCreationValidationError(
                f"TopK question must have min selections greater than 0"
            )
        return self


class QuestionTopKEnhanced(QuestionCheckBoxEnhanced):
    """
    Same as a CheckBox question except required selections count is fixed.
    """

    question_type = "top_k"

    def __init__(self, question: BaseModel):
        # Set the required number of option selections
        min_selections: int
        max_selections: int

        super().__init__(question)

    def construct_answer_data_model(self) -> Type[BaseModel]:
        return super().construct_answer_data_model()



import re
import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Optional, Type
from edsl.questions.experimental.QuestionShortAnswer import QuestionShortAnswer
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings
from edsl.utilities.utilities import random_string


class QuestionShortAnswerText(QuestionShortAnswer):
    question_type: str = "short_answer_text"

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked the following question requiring a text response: 
        {{question_text}}
        {% if expression_constraint %}
        Your answer must satisfy the following constraints (expressed in Python):
        {{expression_constraint}}
        {% endif %}
        Return a valid JSON formatted exactly as follows that satisfies the constraints: 
        {"answer": "<put text answer here>", "comment": "<put explanation here>"}  

        For example, for the question is "What is your job?" with constraints 
        len(answer) < 25 and 'job' not in answer
        the following response would be valid:
        {"answer": "I work as a locksmith", "comment": "My work involves keys and locks."}
        """
        )

    def __repr__(self):
        return f"""{self.__class__.__name__}(question_text = "{self.question_text}", expression_constraints = {self.expression_constraints}, question_name = "{self.question_name}")"""

    @classmethod
    def construct_question_data_model(cls) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionShortAnswerText"""

            question_text: str = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            expression_constraint: str = Field(
                ..., min_length=0, max_length=Settings.MAX_EXPRESSION_CONSTRAINT_LENGTH
            )
            allow_nonresponse: Optional[bool] = None

            @field_validator("expression_constraint")
            def check_expression_constraint_logic(cls, value):
                if not re.match(
                    r"^[a-zA-Z0-9_\s\'\.\+\-\*\/\(\)\%\>\<\=\!\&\|\~\^\,\:\;\[\]\{\}]*$",
                    value,
                ):
                    raise Exception(
                        "Expression constraint contains invalid characters."
                    )
                return value

        return LocalQuestionData

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: str = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise ValueError("You must provide a response.")
                return value

            @field_validator("answer")
            def check_expression_constraint(cls, value):
                print(value)
                formatted_value = (
                    "'" + value + "'"
                )  # add quotes to make it insertable with quotes
                if not eval(
                    self.expression_constraint.replace("answer", formatted_value)
                ):
                    # if not simple_eval(self.expression_constraint, names={'answer': formatted_value}):
                    raise Exception("Answer does not match expression constraint.")
                return value

        return LocalAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        return {"answer": random_string()}

    def form_elements(self):
        raise NotImplementedError


if __name__ == "__main__":
    q = QuestionShortAnswerText(
        question_text="What would you like to be when you grow up?",
        expression_constraint="len(answer) < 50 and 'imagine' in answer",
        question_name="grow_up",
    )
    from edsl import Agent, Scenario, print_dict_with_rich

    a_20yo_finance_student = Agent(
        traits={"age": "20 years old", "occupation": "finance student"}
    )
    print_dict_with_rich(a_20yo_finance_student.answer_question(q))



import random
import textwrap
from jinja2 import Template
from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Optional, Type
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.exceptions import (
    QuestionCreationValidationError,
    QuestionAnswerValidationError,
)
from edsl.utilities.utilities import random_string


class QuestionCheckBox(QuestionData):
    """Pydantic data model for QuestionCheckBox"""

    question_options: list[str] = Field(
        ...,
        min_length=Settings.MIN_NUM_OPTIONS,
        max_length=Settings.MAX_NUM_OPTIONS,
    )
    min_selections: Optional[int] = None
    max_selections: Optional[int] = None

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionCheckBoxEnhanced":
        instance = super(QuestionCheckBox, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionCheckBoxEnhanced(instance)

    @model_validator(mode="after")
    def check_options_count(self):
        if (
            hasattr(self, "question_options")
            and hasattr(self, "min_selections")
            and self.min_selections != None
        ):
            if self.min_selections > len(self.question_options):
                raise QuestionCreationValidationError(
                    f"You asked for at least {self.min_selections} selections, but provided {len(self.question_options)} options."
                )
        if (
            hasattr(self, "question_options")
            and hasattr(self, "max_selections")
            and self.max_selections != None
        ):
            if self.max_selections > len(self.question_options):
                raise QuestionCreationValidationError(
                    f"You asked for at most {self.max_selections} selections, but provided {len(self.question_options)} options."
                )
        return self

    @field_validator("question_options")
    def check_unique(cls, value):
        return cls.base_validator_check_unique(value)

    @field_validator("question_options")
    def check_option_string_lengths(cls, value):
        return cls.base_validator_check_option_string_lengths(value)


class QuestionCheckBoxEnhanced(Question):
    question_type = "checkbox"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        # make the min/max conditional parts of the instructions {% if ... end if}
        # if no min_value then default = 1 option must be selected
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        The options are 
        {% for option in question_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        Return a valid JSON formatted like this, selecting only the number of the option: 
        {"answer": [<put comma-separated list of answer codes here>], "comment": "<put explanation here>"}
        {% if min_selections != None and max_selections != None and min_selections == max_selections %}
        You must select exactly {{min_selections}} options.
        {% elif min_selections != None and max_selections != None %}
        Minimum number of options that must be selected: {{min_selections}}.      
        Maximum number of options that must be selected: {{max_selections}}.
        {% elif min_selections != None %}
        Minimum number of options that must be selected: {{min_selections}}.      
        {% elif max_selections != None %}
        Maximum number of options that must be selected: {{max_selections}}.      
        {% endif %}        
        """
        )

    def construct_answer_data_model(self) -> Type[BaseModel]:
        acceptable_values = range(len(self.question_options))

        class QuestionCheckBoxAnswerDataModel(AnswerData):
            answer: list[int]

            @field_validator("answer")
            def check_answers_valid(cls, value):
                if all([v in acceptable_values for v in value]):
                    return value
                else:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} has elements not in {acceptable_values}."
                    )

            @field_validator("answer")
            def check_answers_count(cls, value):
                # If min or max numbers of option selections are specified, check they are satisfied
                if self.min_selections is not None and len(value) < self.min_selections:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} has fewer than {self.min_selections} options selected."
                    )
                if self.max_selections is not None and len(value) > self.max_selections:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} has more than {self.max_selections} options selected."
                    )
                return value

        return QuestionCheckBoxAnswerDataModel

    ################
    # Less important
    ################
    def translate_answer_code_to_answer(self, answer_codes, scenario):
        """
        Translates the answer code to the actual answer.
        For example, for question options ["a", "b", "c"],the answer codes are 0, 1, and 2.
        The LLM will respond with [0,1] and this code will translate it to ["a","b"].
        """
        if scenario is None:
            scenario = dict()
        translated_options = [
            Template(option).render(scenario) for option in self.question_options
        ]

        translated_codes = []
        for answer_code in answer_codes:
            translated_codes.append(translated_options[int(answer_code)])
        return translated_codes

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            answer = {
                "answer": [random.choice(self.question_options)],
                "comment": random_string(),
            }
        else:
            answer = {
                "answer": [random.choice(range(len(self.question_options)))],
                "comment": random_string(),
            }
        return answer

    def form_elements(self):
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        for index, option in enumerate(self.question_options):
            html_output += f"""<div id = "{self.question_name}_div_{index}">
            <input type="checkbox" id="{self.question_name}_{index}" name="{self.question_name}" value="{option}">
            <label for="{self.question_name}_{index}">{option}</label>
            </div>\n"""
        return html_output



import keyword
from pydantic import BaseModel, ConfigDict, Field, field_validator
from typing import Optional
from edsl.exceptions import QuestionCreationValidationError
from edsl.questions.settings import Settings
from edsl.utilities.utilities import random_string


class QuestionData(BaseModel):
    """
    Pydantic model that all question-specific question pydantic models inherit from.
    Used to put commonly used validators in one place.
    """

    question_name: Optional[str] = Field(
        None, min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )

    short_names_dict: Optional[dict[str, str]] = Field(default_factory=dict)

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )

    @field_validator("question_name")
    def check_name(cls, value):
        return cls.base_validator_check_question_name(value)

    # forbid attributes not defined in the model
    # accept attributes with types that the model doesn't know how to validate
    model_config = ConfigDict(extra="forbid", arbitrary_types_allowed=True)

    # Place commonly used validators below
    def base_validator_check_option_string_lengths(value: list[str]) -> list[str]:
        """Check that each option string is not too long."""
        limit = Settings.MAX_OPTION_LENGTH
        for option in value:
            length = len(option)
            if length > limit:
                raise QuestionCreationValidationError(
                    f"Option {option} is too long ({length} chars, limit={limit} chars)"
                )
            elif length == 0:
                raise QuestionCreationValidationError(
                    f"Option {option} cannot be empty."
                )
        return value

    def base_validator_check_question_name(value: Optional[str]) -> str:
        """Check that question_name is a valid Python identifier and not a Python keyword."""
        # if value is None, generate a random string
        if value is None:
            value = random_string()
        # check that the Question name is a valid Python identifier
        if not value.isidentifier():
            raise QuestionCreationValidationError(
                f"`question_name` must be a valid identifier, but you passed in {value}"
            )
        # check that the Question name is not a Python keyword
        if value in keyword.kwlist:
            raise QuestionCreationValidationError(
                f"`question_name` cannot be a Python keyword, but you passed in {value}"
            )
        return value

    def base_validator_check_unique(value: list[str]) -> list[str]:
        """Check that all elements in a list are unique."""
        if len(value) != len(set(value)):
            raise QuestionCreationValidationError(
                f"Elements must be unique, but you passed in {value}"
            )
        return value


class AnswerData(BaseModel):
    """
    Pydantic model that all question-specific answer pydantic models inherit from.
    Used to put commonly used validators in one place.
    """

    # Place commonly used validators below
    # def base_validator_lte(value, max_value: int) -> int:
    #     """Check that value is less than or equal to max_value."""
    #     if value > max_value:
    #         raise ValueError(f"Value {value} is greater than {max_value}")
    #     return value



from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Type, Union

from edsl.questions.Question import Question
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings


class QuestionMatrix(Question):
    question_type = "matrix"

    """This is a question where the answer is a matrix.
    It looks like this:

            | Yuk   | Yum   |
    --------+-------+-------+
    Hot Dog |   X   |       |
    --------+-------+-------+
    Pizza   |       |   X   |
    --------+-------+-------+

    You have to select one option from each row and column if answer_required.
    Otherise, you can leave a row blank.
    """

    def construct_question_data_model(self) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionMatrix"""

            question_text: str = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            row_options: list[str]
            column_options: list[str]
            answer_required: bool = False

            @field_validator("row_options")
            def check_unique(cls, value):
                return cls.base_validator_check_unique(value)

            @field_validator("row_options")
            def check_option_string_lengths(cls, value):
                return cls.base_validator_check_option_string_lengths(value)

            @field_validator("column_options")
            def check_unique(cls, value):
                return cls.base_validator_check_unique(value)

            @field_validator("column_options")
            def check_option_string_lengths(cls, value):
                return cls.base_validator_check_option_string_lengths(value)

            # raise NotImplementedError("I'm getting tired of these")

        return LocalQuestionData

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: list[Union[int, None]]

            @model_validator(mode="after")
            def check_answer(self):
                if self.answer_required:
                    for row in self.answer.keys():
                        for column in self.answer[row].keys():
                            if self.answer[row][column] is None:
                                raise ValueError(
                                    f"Answer required but not provided for {row}, {column}"
                                )
                return self

            # raise NotImplementedError("I'm getting tired of these")

        return LocalAnswerDataModel



from pydantic import BaseModel, Field
from typing import Type
from edsl.questions import Settings, QuestionData
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoiceEnhanced


class QuestionLikertFive(QuestionData):
    """Pydantic data model for QuestionLikertFive"""

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )
    question_options: list[str] = [
        "Strongly disagree",
        "Disagree",
        "Neutral",
        "Agree",
        "Strongly agree",
    ]

    def __new__(cls, *args, **kwargs):
        instance = super(QuestionLikertFive, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionLikertFiveEnhanced(instance)


class QuestionLikertFiveEnhanced(QuestionMultipleChoiceEnhanced):
    """
    Inherits from QuestionMultipleChoice, because the two are similar.
    - A difference is that the answers in QuestionLikertFive are fixed
      and have a very specific order and labels
    """

    question_type = "likert_five"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    def construct_answer_data_model(self) -> Type[BaseModel]:
        """Reuses the answer data model from QuestionMultipleChoiceEnhanced"""
        return super().construct_answer_data_model()

    # TODO: Seems that we need to have the label of each option?
    def form_elements(self):
        scale_values = ["1", "2", "3", "4", "5"]
        html_output = f"""
        <label>{self.question_text}</label>\n"""

        for index, value in enumerate(scale_values):
            html_output += f"""
            <div id="{self.question_name}_div_{index}">
                <input type="radio" id="{self.question_name}_{index}" 
                    name="{self.question_name}" value="{value}">
                <label for="{self.question_name}_{index}">{value}</label>
            </div>\n"""

        return html_output



from typing import Type, Optional
from pydantic import BaseModel, Field, field_validator, model_validator
from edsl.exceptions import (
    QuestionAnswerValidationError,
    QuestionCreationValidationError,
)
from edsl.questions import Settings, QuestionData, AnswerData
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoiceEnhanced


class QuestionLinearScale(QuestionData):
    """Pydantic data model for QuestionLinearScale"""

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )
    question_options: list[int]
    option_labels: Optional[dict[int, str]] = None

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionLinearScaleEnhanced":
        instance = super(QuestionLinearScale, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionLinearScaleEnhanced(instance)

    @field_validator("question_options")
    def check_successive(cls, value):
        if sorted(value) != list(range(min(value), max(value) + 1)):
            raise QuestionCreationValidationError(
                f"LinearScale.question_options must be a list of successive integers, e.g. [1, 2, 3]"
            )
        return value

    @model_validator(mode="after")
    def check_option_labels(self, value):
        if self.option_labels is not None:
            if min(self.option_labels.keys()) != min(self.question_options):
                raise QuestionCreationValidationError(f"First option needs a label")
            if max(self.option_labels.keys()) != max(self.question_options):
                raise QuestionCreationValidationError(f"Last option needs a label")
        return self


class QuestionLinearScaleEnhanced(QuestionMultipleChoiceEnhanced):
    """
    Inherits from QuestionMultipleChoice, because the two are similar.
    - A difference is that the answers must have an ordering.
    - Not every option has to have a label.
    - But if option labels are provided, there have to be labels for the first and last options.
    """

    question_type = "linear_scale"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    def construct_answer_data_model(self) -> Type[BaseModel]:
        "Constructs the answer data model for this question"
        acceptable_values = self.question_options

        class QuestionLinearScaleAnswerDataModel(AnswerData):
            answer: int
            comment: Optional[str] = None

            @field_validator("answer")
            def check_answer(cls, value):
                if value in acceptable_values:
                    return value
                else:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} is not in the acceptable values {acceptable_values}"
                    )

        return QuestionLinearScaleAnswerDataModel

    def form_elements(self) -> str:
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        for index in range(1, 6):  # assuming the scale is from 1 to 5
            html_output += f"""<div id = "{self.question_name}_div_{index}">
            <input type="radio" id="{self.question_name}_{index}" name="{self.question_name}" value="{index}">
            <label for="{self.question_name}_{index}">{index}</label>
            </div>\n"""
        return html_output



from pydantic import BaseModel
from typing import Optional, Type, Callable
from edsl.questions import Question, QuestionData


class QuestionFunctional(QuestionData):
    """Pydantic data model for QuestionFunctional"""

    question_name: Optional[str] = None
    func: Callable

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionFunctionalEnhanced":
        instance = super(QuestionFunctional, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionFunctionalEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)


class QuestionFunctionalEnhanced(Question):
    """
    A special type of question that's *not* answered by an LLM
    - Instead, it is "answered" by a function that is passed in, `func`.
    - Useful for questions that require some kind of computation first
      or are the result of a multi-step process.
    See `compose_questions` in `compose_functions.py` for an example of how this is used.

    Notes
    - 'func' is a function that takes in a scenario and agent traits and returns an answer.
    - QuestionFunctional is not meant to be instantiated directly by end-users, but rather
      it is meant to be subclassed by us to create new function types.
    - It is probably *not* safe to allow end-users to have the ability to pass functional-derived questions.
      They could monkey-patch the function to do something malicious, e.g., to replace our function logic
      with "os.system('rm -rf /')".
    - One possible solution is to have interfaces they can pass via the API, like so:
      QuestionDropdown(question_name = "dropdown", question_options = ["a", "b", "c"]...)
      which we then translate to the real QuestionFunctional `under the hood.`

    To see how it's used, see `tests/test_QuestionFunctional_construction_from_function`
    """

    question_type = "functional"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self):
        """Required by Question, but not used by QuestionFunctional"""
        return None

    def answer_question_directly(self, scenario, agent_traits=None):
        return {"answer": self.func(scenario, agent_traits), "comment": None}

    def translate_answer_code_to_answer(self, answer, scenario):
        """Required by Question, but not used by QuestionFunctional"""
        return None

    def construct_answer_data_model(self) -> Type[BaseModel]:
        """Required by Question, but not used by QuestionFunctional"""
        return None

    def form_elements(self) -> str:
        """Required by Question, but not used by QuestionFunctional"""
        raise NotImplementedError

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Required by Question, but not used by QuestionFunctional"""
        raise NotImplementedError



import os
from edsl import BASE_DIR


EXAMPLE_RESULTS_PATH = os.path.join(BASE_DIR, "utilities/data/example_results.json")
SCOOTER_RESULTS_PATH = os.path.join(BASE_DIR, "utilities/data/scooter_results.json")



import random
import textwrap
from pydantic import BaseModel, model_validator
from typing import Any, Type
from edsl.questions import Question, QuestionData, AnswerData
from edsl.exceptions import QuestionAnswerValidationError
from edsl.utilities.utilities import random_string


class QuestionExtract(QuestionData):
    """Pydantic data model for QuestionExtract"""

    answer_template: dict[str, Any]

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionExtractEnhanced":
        instance = super(QuestionExtract, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionExtractEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)


class QuestionExtractEnhanced(Question):
    question_type = "extract"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        return textwrap.dedent(
            """\
        You are given the following input: "{{question_text}}".
        Create an ANSWER should be formatted like this {{ answer_template }},
        and it should have the same keys but values extracted from the input.
        If the value of an key is not present in the input, fill with "null".
        Return a valid JSON formatted like this: 
        {"answer": <put your ANSWER here>}
        ONLY RETURN THE JSON, AND NOTHING ELSE.
        """
        )

    def construct_answer_data_model(self) -> Type[BaseModel]:
        "Constructs the answer data model for this question"
        acceptable_answer_keys = set(self.answer_template.keys())

        class QuestionExtractAnswerDataModel(AnswerData):
            answer: dict

            @model_validator(mode="after")
            def check_answer(self):
                if any(
                    [key not in acceptable_answer_keys for key in self.answer.keys()]
                ):
                    raise QuestionAnswerValidationError(
                        f"Answer keys must be in {acceptable_answer_keys}, but got {self.answer.keys()}"
                    )
                if any(
                    [key not in self.answer.keys() for key in acceptable_answer_keys]
                ):
                    raise QuestionAnswerValidationError(
                        f"Answer must have all keys in {acceptable_answer_keys}, but got {self.answer.keys()}"
                    )

        return QuestionExtractAnswerDataModel

    ################
    # Less important
    ################
    # TODO - ALL THE BELOW
    def translate_answer_code_to_answer(self, answer, scenario):
        return answer

    def simulate_answer(self, human_readable=True) -> dict[str, str]:
        """Simulates a valid answer for debugging purposes"""
        if human_readable:
            answer = random.choice(self.question_options)
        else:
            answer = random.choice(range(len(self.question_options)))
        return {
            "answer": answer,
            "comment": random_string(),
        }

    def form_elements(self) -> str:
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        return html_output


# main
if __name__ == "__main__":
    from edsl.questions import QuestionExtract

    question = QuestionExtract(
        question_text="My name is Moby Dick. I have a PhD in astrology, but I'm actually a truck driver",
        answer_template={"name": "John Doe", "profession": "Carpenter"},
        question_name="extract_name",
    )
    question.run()

    question = QuestionExtract(
        question_text="Lorem Ipsum is simply dummy text of the printing and typesetting industry. Lorem Ipsum has been the industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type and scrambled it to make a type specimen book. It has survived not only five centuries, but also the leap into electronic typesetting, remaining essentially unchanged. It was popularised in the 1960s with the release of Letraset sheets containing Lorem Ipsum passages, and more recently with desktop publishing software like Aldus PageMaker including versions of Lorem Ipsum.",
        answer_template={"name": "John Doe", "profession": "Carpenter"},
        question_name="extract_name",
    )
    question.run()

    question = QuestionExtract(
        question_text="Please be advised that your timecard(s) are due to be submitted by 11:30 PM Sunday. Please log in to the time keeping portal to process your timecard. Thank you. Click Here to Login. Click Here to Login on your Mobile Device. Note: Please do not reply to this message, as it is system generated.",
        answer_template={"time_critical": "true", "system_generated": "false"},
        question_name="extract_email",
    )
    question.run()

    question = QuestionExtract(
        question_text="My name is Sam and I was born on 1992.",
        answer_template={"name": "John Doe", "age": 33},
        question_name="extract_name_age",
    )
    question.run()



"""
Results is a UserList with some special methodsBarChart.

It is instantiated with a survey and a list of observations.
The first time called, it copies over that list of observations to self.original_data.

The Results object can then be manipulated in various ways with select, filter, apply, etc. 
In each case, the original data is preserved because it is passed to the __init__ method 
of Results along with the new list of data. 
This way you can pass the results to, say, histogram or wordcloud and it can use "self" as the input. 

Regardless, it can always be restored to its original state with the method ".restore()".

## Goals: 
- filter & where can work with the full list 
- SQL-like, it will make the right inference if the 'table' is missing

"""
from __future__ import annotations
import copy
import io
import jinja2
import json
import markdown2
import sys
import textwrap
from typing import Optional, Union
from collections import UserList, defaultdict
from simpleeval import EvalWithCompoundTypes
from edsl.utilities.interface import (
    print_list_of_dicts_with_rich,
    print_list_of_dicts_as_html_table,
)
from edsl.results.Result import Result
from edsl.utilities.utilities import shorten_string
from edsl.results.ResultsExportMixin import ResultsExportMixin
from edsl.results.RegressionMixin import RegressionMixin
from edsl.results.ResultsOutputMixin import ResultsOutputMixin
from edsl.results.ResultsFetchMixin import ResultsFetchMixin
from edsl.utilities.utilities import is_gzipped


class ResultsStates:
    SELECTED = "selected"
    ORIGINAL = "original"


class Results(
    UserList,
    RegressionMixin,
    ResultsExportMixin,
    ResultsOutputMixin,
    ResultsFetchMixin,
):
    def __init__(
        self,
        survey,
        data,
        original_data=None,
        filtered_data=None,
        report_items=None,
        created_columns=None,
        state=ResultsStates.ORIGINAL,
    ):
        """
        -- self.data - working list
        -- self.original_data - what we started with
        ---self.filtered_data - what we have filtered things down too

        -- If you select a single column, it should just return a list.
        -- If you select multiple columns, it should return a list of dictionaries.
        """
        super().__init__(data)
        self.survey = survey
        self.state = state

        if original_data is None:  # instantiated for the first time
            self.original_data = copy.copy(data)
            self.filtered_data = copy.copy(data)
            self.report_items = []
            self.created_columns = []
        else:
            self.original_data = original_data
            self.filtered_data = filtered_data
            self.report_items = report_items
            self.created_columns = created_columns

        ResultsOutputMixin.add_output_functions(self)

    @property
    def _markdown_reprsentation(self):
        template = jinja2.Template(
            textwrap.dedent(
                """\
            # Results Report
            ## Overview
            - Observations: {{ results | length }}
            - Survey questions: {{ results.survey | length }}

            ## Keys
            - Answer keys: `{{ results.answer_keys }}`
            - Agent trait keys: `{{ results.agent_keys }}`
            - Scenario keys: `{{ results.scenario_keys }}`
            - Model keys: `{{ results.model_keys }}`
            - Created columns: `{{ results.created_columns }}`

            ## Survey questions
            {% for question in questions: %}
            1. (`{{ question.question_name }}`) {{ question.question_text }}
            {% if question.question_type == "multiple_choice": %}
                Type: {{ question.question_type }}
            {% for option in question.question_options %}
                    - {{ option }}
            {% endfor %}
            {% elif question.question_type == "free_text": %}
                Type: {{ question.question_type }}
            {% endif %}
            {% endfor %}
            """
            )
        )
        rendered_markdown = template.render(
            results=self, questions=self.survey._questions
        )
        return rendered_markdown

    def _repr_html_(self):
        rendered_markdown = self._markdown_reprsentation
        rendered_html = markdown2.markdown(rendered_markdown)
        return rendered_html

    def __repr__(self):
        return self._markdown_reprsentation

    @property
    def key_to_data_type(self):
        print("Switch to using _key_to_data_type. This shouldn't be called externally")
        return self._key_to_data_type

    @property
    def _key_to_data_type(self):
        """E.g.,
        {'temperature': 'model',
        'max_tokens': 'model',
        'top_p': 'model',
        'frequency_penalty': 'model',
        'presence_penalty': 'model',
        'use_cache': 'model',
        'model': 'model',
        'baseline': 'result'}

        >>> results = Results.create_example()
        >>> d = results._key_to_data_type
        >>> mapping = {'status': 'agent', 'period': 'scenario', 'temperature': 'model', 'max_tokens': 'model', 'top_p': 'model', 'frequency_penalty': 'model', 'presence_penalty': 'model', 'use_cache': 'model', 'model': 'model', 'how_feeling': 'answer', 'elapsed': 'answer'}
        >>> d == mapping
        True
        """
        d = {}
        for result in self.original_data:
            d.update(result.key_to_data_type)

        # The user could have created columns using 'mutate'
        for column in self.created_columns:
            d[column] = "answer"

        return d

    @property
    def data_type_to_keys(self):
        print("Switch to using _data_type_to_keys. This shouldn't be called externally")
        return self._data_type_to_keys

    @property
    def _data_type_to_keys(self) -> dict:
        """Maps data types (result, agent, scenario, etc.) to keys (how_feeling, status, etc.)

        >>> r = Results.create_example()
        >>> mapping = r._data_type_to_keys.keys()
        """
        d = defaultdict(set)
        for result in self.original_data:
            for key, value in result.key_to_data_type.items():
                d[value] = d[value].union(set({key}))

        for column in self.created_columns:
            d["answer"] = d["answer"].union(set({column}))
        return d

    @property
    def answer_keys(self) -> dict:
        # return self._data_type_to_keys["answer"]
        answer_keys = self._data_type_to_keys["answer"]
        answer_keys = {k for k in answer_keys if "_comment" not in k}
        questions_text = [
            self.survey.get_question(k).question_text for k in answer_keys
        ]
        short_question_text = [shorten_string(q, 80) for q in questions_text]
        return dict(zip(answer_keys, short_question_text))
        return {
            "survey_questions": self._data_type_to_keys["answer"],
            "created": self.created_columns,
        }

    @property
    def agent_keys(self):
        return self._data_type_to_keys["agent"]

    @property
    def scenario_keys(self):
        return self._data_type_to_keys["scenario"]

    @property
    def model_keys(self):
        return self._data_type_to_keys["model"]

    @property
    def all_keys(self):
        answer_keys = set(self.answer_keys)
        return (
            answer_keys.union(self.agent_keys)
            .union(self.scenario_keys)
            .union(self.model_keys)
        )

    def restore(self) -> None:
        """Restore the data to its original state.

        >>> r = Results.create_example()
        >>> r2 = r.filter("how_feeling == 'Great'").select('how_feeling')
        >>> r = Results.create_example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling')
        [{'answer.how_feeling': ['Great', 'Great']}]

        Restore the data to its original state.
        >>> r = Results.create_example()
        >>> manipulated_data = r.filter("how_feeling == 'Great'").select('how_feeling')
        >>> r.restore()
        >>> r == manipulated_data
        False
        """
        self.data = self.original_data
        self.filtered_data = self.original_data
        self.state = ResultsStates.ORIGINAL

    def parse_column(self, column: str) -> tuple[str, str]:
        print("Change to using _parse_column here")
        return self._parse_column(column)

    def _parse_column(self, column: str) -> tuple[str, str]:
        """
        Utility function to parse a column name into a data type and a key.
        The standard way a column is specified is with a dot-separated string, e.g. "agent.status".

        But you can also specify a single key, e.g. "status", in which case it will look up the data type.
        This relies on the key_to_data_type property of the Results class.
        """
        if "." in column:
            data_type, key = column.split(".")
        else:
            if column in self._key_to_data_type:
                data_type = self._key_to_data_type[column]
                key = column
            else:
                breakpoint()
                raise Exception(f"Column {column} not found in data")

        return data_type, key

    def first(self):
        """
        This returns the first observation in the results.
        >>> r = Results.create_example()
        >>> r.select('how_feeling').first()
        'Bad'
        """
        return list(self[0].values())[0][0]

    def mutate(
        self, new_var_string, functions_dict=None, in_place=True
    ) -> Union[Results, None]:
        """
        Creates a value value in 'results' as if has been asked as part of the survey.
        It splits the new_var_string at the "=" and uses simple_eval

        >>> r = Results.create_example()
        >>> r.mutate('how_feeling_x = how_feeling + "x"').select('how_feeling_x')
        [{'answer.how_feeling_x': ['Badx', 'Badx', 'Greatx', 'Greatx']}]

        TODO: Make sure left is valid python name
        TODO: Do something smarter than serialize/deserialize the Result
        """
        raw_var_name, expression = new_var_string.split("=", 1)

        var_name = raw_var_name.strip()
        new_data = []
        for result in self.filtered_data:
            try:
                # the combined_dict is a Result-level dictionary mapping keys to values
                # e.g., {'agent': {'status':'OK'}} as well as {'status':'OK'}
                # This can deal w/ the case whree the user uses dot notation
                if functions_dict is not None:
                    evaluator = EvalWithCompoundTypes(
                        names=result.combined_dict, functions=functions_dict
                    )
                else:
                    evaluator = EvalWithCompoundTypes(names=result.combined_dict)

                value = evaluator.eval(expression)

                # this bit of weirdness is done because we need a deep copy
                # but because result is an iterator, it's got a lock on it (i think)

                json_string = json.dumps(result.to_dict())

                ## serialization is noisy so we're going to suppress it
                original_stdout = (
                    sys.stdout
                )  # Save a reference to the original standard output
                with io.StringIO() as buffer:
                    sys.stdout = buffer  # Redirect stdout to buffer
                    new_result = Result.from_dict(json.loads(json_string))
                    captured_output = (
                        buffer.getvalue()
                    )  # Get printed content from buffer

                sys.stdout = original_stdout  # Restore original stdout

                new_result["answer"][var_name] = value
                new_data.append(new_result)
            except Exception as e:
                print(f"Exception:{e}")

        if in_place:
            self.data = new_data
            self.filtered_data = new_data
            self.created_columns = self.created_columns + [var_name]
            return None
        else:
            new_results = Results(
                survey=self.survey,
                data=new_data,
                original_data=self.original_data,
                filtered_data=new_data,
                state=self.state,
                created_columns=self.created_columns + [var_name],
            )

            # if it was in a selected state, keep it that way
            if self.state == ResultsStates.SELECTED:
                relevent_columns = self.relevant_columns()
                new_results = new_results.select(*relevent_columns)

            return new_results

    def select(self, *columns: Union[str, list[str]]) -> Results:
        """
        This selects data from the results and turns it into a format like so:
        >>> results = Results.create_example()
        >>> results.select('how_feeling')
        [{'answer.how_feeling': ['Bad', 'Bad', 'Great', 'Great']}]
        """
        # We start by changing the data back to format where self.data is a list of Result
        # objects. However, it can be filtered, so we need to use self.filtered_data.

        self.data = self.filtered_data

        # we're doing to populate this with the data we want to fetch
        to_fetch = defaultdict(list)

        if not columns or columns == ("*",) or columns == (None,):
            columns = ("*.*",)

        if isinstance(columns[0], list):
            columns = tuple(columns[0])

        new_data = []
        items_in_order = []
        for column in columns:
            # a user could pass 'result.how_feeling' or just 'how_feeling'
            parsed_data_type, parsed_key = self._parse_column(column)
            if parsed_data_type == "*":
                data_types = ["answer", "scenario", "agent", "model"]
            else:
                if parsed_data_type not in ["answer", "scenario", "agent", "model"]:
                    raise Exception(f"Data type {parsed_data_type} not found in data")
                data_types = [parsed_data_type]

            found_once = False  # we need to track this to make sure we found the key at least once
            for data_type in data_types:
                relevant_keys = self._data_type_to_keys[data_type]
                if (
                    parsed_key == "*"
                ):  # wildcard, we'll get all the keys for that data type
                    found_once = True
                    to_fetch[data_type] = relevant_keys
                    items_in_order.extend([data_type + "." + k for k in relevant_keys])
                else:
                    for key in relevant_keys:
                        if key == parsed_key:
                            found_once = True
                            to_fetch[data_type].append(key)
                            items_in_order.append(data_type + "." + key)

            if not found_once:
                raise Exception(f"Key {parsed_key} not found in data")

        for data_type in to_fetch:
            for key in to_fetch[data_type]:
                entries = self._fetch_list(data_type, key)
                new_data.append({data_type + "." + key: entries})

        def sort_by_key_order(dictionary):
            # Extract the single key from the dictionary
            single_key = next(iter(dictionary))
            # Return the index of this key in the list_of_keys
            return items_in_order.index(single_key)

        sorted(new_data, key=sort_by_key_order)

        # print(f"Items in order: {items_in_order}")
        new_results = Results(
            survey=self.survey,
            data=sorted(new_data, key=sort_by_key_order),
            filtered_data=self.filtered_data,
            original_data=self.original_data,
            report_items=self.report_items,
            created_columns=self.created_columns,
            state=ResultsStates.SELECTED,
        )
        return new_results

    def print(
        self,
        pretty_labels=None,
        drop=None,
        sort_by=None,
        reverse=True,
        filename=None,
        html=False,
        interactive=False,
        split_at_dot=True,
    ):
        """
        This prints the results in a nice format. It's not fluent.
        """
        if sort_by is not None:
            self = self.sort_by(sort_by, reverse=reverse)

        if pretty_labels is None:
            pretty_labels = {}

        if self.state == ResultsStates.SELECTED:
            new_data = []
            for entry in self:
                key, list_of_values = list(entry.items())[0]
                new_data.append({pretty_labels.get(key, key): list_of_values})
            else:
                if not html:
                    print_list_of_dicts_with_rich(
                        new_data, filename=filename, split_at_dot=split_at_dot
                    )
                else:
                    print_list_of_dicts_as_html_table(
                        new_data, filename=None, interactive=interactive
                    )

        elif self.state == ResultsStates.ORIGINAL:
            selected_self = self.select()
            selected_self.print(
                pretty_labels=pretty_labels,
                drop=drop,
                sort_by=sort_by,
                reverse=reverse,
                filename=filename,
            )

    def sort_by(self, column, reverse=True) -> Results:
        "Sorts the results by a column"

        data_type, key = self._parse_column(column)

        def to_numeric_if_possible(v):
            try:
                return float(v)
            except:
                return v

        if self.state == ResultsStates.ORIGINAL:
            new_data = sorted(
                self.data,
                key=lambda x: to_numeric_if_possible(x.get_value(data_type, key)),
                reverse=reverse,
            )
            new_filtered_data = sorted(
                self.filtered_data,
                key=lambda x: to_numeric_if_possible(x.get_value(data_type, key)),
                reverse=reverse,
            )
            new_results = Results(
                survey=self.survey,
                data=new_data,
                original_data=self.original_data,
                filtered_data=new_filtered_data,
                report_items=self.report_items,
                created_columns=self.created_columns,
                state=self.state,
            )
            return new_results
        elif self.state == ResultsStates.SELECTED:
            relevant_columns = self.relevant_columns()
            if len(self.filtered_data) < len(self.original_data):
                print("Warning: sorting on a filtered dataset removes filters")
            self.restore()
            return self.sort_by(column, reverse=reverse).select(*relevant_columns)
        else:
            raise Exception("Cannot sort on a tabular dataset yet")

    def filter(self, expression) -> Results:
        """
        This filters a result based on the expression that is passed in.
        >>> r = Results.create_example()
        >>> r.filter("how_feeling == 'Great'").select('how_feeling')
        [{'answer.how_feeling': ['Great', 'Great']}]
        >>> r.filter("how_feeling == 'Nothing'").select('how_feeling')
        [{'answer.how_feeling': []}]
        """
        new_data = []
        for result in self.filtered_data:
            try:
                # the combined_dict is a Result-level dictionary mapping keys to values
                # e.g., {'agent': {'status':'OK'}} as well as {'status':'OK'}
                # This can deal w/ the case whree the user uses dot notation
                evaluator = EvalWithCompoundTypes(names=result.combined_dict)
                if eval_result := evaluator.eval(expression):
                    new_data.append(result)
            except Exception as e:
                print(f"Exception:{e}")

        new_results = Results(
            survey=self.survey,
            data=new_data,
            original_data=self.original_data,
            filtered_data=new_data,
            report_items=self.report_items,
            created_columns=self.created_columns,
            state=self.state,
        )

        # if it was in a selected state, keep it that way
        if self.state == ResultsStates.SELECTED:
            relevent_columns = self.relevant_columns()
            new_results = new_results.select(*relevent_columns)

        return new_results

    def relevant_columns(self) -> set:
        """
        >>> r = Results.create_example()
        >>> keys = r.relevant_columns()
        >>> keys == {'elapsed', 'temperature', 'agent', 'scenario', 'answer', 'use_cache', 'how_feeling', 'frequency_penalty', 'max_tokens', 'model', 'status', 'top_p', 'period', 'presence_penalty'}
        True
        >>> r.select('how_feeling').relevant_columns()
        {'answer.how_feeling'}
        """
        if self.state == ResultsStates.SELECTED:
            keys = set([list(d.keys())[0] for d in self.data])
        else:
            keys = set({})
            for observation in self:
                observation_keys = set(observation.combined_dict.keys())
                keys = keys.union(observation_keys)
        return keys

    def to_dict(self):
        self.restore()  # restore to original state before converting to dict
        new_data = []
        for observation in self:
            new_observation = observation.to_dict()
            new_data.append(new_observation)
        return {
            "data": new_data,
            "survey": self.survey.to_dict(),
            "agents": [a.to_dict() for a in self.agents],
        }

    @classmethod
    def from_dict(cls, data: dict):
        from edsl.surveys.Survey import Survey
        from edsl.results.Result import Result
        from edsl.agents import Agent

        survey = Survey.from_dict(data["survey"])
        # agents = [Agent.from_dict(a) for a in data["agents"]]
        observations = [Result.from_dict(r) for r in data["data"]]
        new_cls = cls(survey=survey, data=observations)
        # new_cls.agents = agents
        return new_cls

    @classmethod
    def load(cls, json_file):
        if is_gzipped(json_file):
            import gzip

            with gzip.open(json_file, "rb") as f:
                data = json.load(f)
        else:
            with open(json_file, "r") as f:
                data = json.load(f)
        return cls.from_dict(data)

    @property
    def question_names(self):
        return list(self.survey.question_names)

    @property
    def agents(self):
        return [r.agent for r in self.original_data]

    @classmethod
    def create_example(cls, refresh=False, debug=False):
        from edsl.utilities.data.Registry import EXAMPLE_RESULTS_PATH

        file_path = EXAMPLE_RESULTS_PATH

        if refresh:
            from edsl.jobs.Jobs import create_example_jobs, Jobs

            j = create_example_jobs()
            r = j.run(n=1, debug=debug)
            with open(file_path, "w") as file:
                json.dump(
                    r.to_dict(), file
                )  # Assuming r has a to_dict method to serialize the results

            return r

        else:
            original_stdout = (
                sys.stdout
            )  # Save a reference to the original standard output

            with io.StringIO() as buffer:
                sys.stdout = buffer  # Redirect stdout to buffer
                results = cls.from_dict(json.load(open(file_path, "r")))
                captured_output = buffer.getvalue()  # Get printed content from buffer

            sys.stdout = original_stdout  # Restore original stdout

            return results

    def show_methods(self):
        public_methods_with_docstrings = [
            (method, getattr(self, method).__doc__)
            for method in dir(self)
            if callable(getattr(self, method)) and not method.startswith("_")
        ]

        return [x[0] for x in public_methods_with_docstrings]


def create_example_results(debug=False, refresh=False) -> Results:
    print("Change to using class method directly")
    return Results.create_example(refresh=refresh, debug=debug)


if __name__ == "__main__":
    results = Results.create_example(refresh=False)

    import doctest

    doctest.testmod()



import csv
from io import StringIO
import io
import pandas as pd


class ResultsStates:
    SELECTED = "selected"
    ORIGINAL = "original"


class ResultsExportMixin:
    def _make_tabular(self, remove_prefix):
        "Helper function that turns the results into a tabular format."
        d = {}
        full_header = list(self.relevant_columns())
        for entry in self:
            key, list_of_values = list(entry.items())[0]
            d[key] = list_of_values
        if remove_prefix:
            header = [h.split(".")[-1] for h in full_header]
        else:
            header = full_header
        num_observations = len(list(self[0].values())[0])
        rows = []
        # rows.append(header)
        for i in range(num_observations):
            row = [d[h][i] for h in full_header]
            rows.append(row)
        # return header, rows
        return header, rows

    def to_csv(self, filename: str = None, remove_prefix=False, download_link=False):
        """
        >>> r = create_example_results()
        >>> r.select('how_feeling').to_csv()
        'result.how_feeling\\r\\nBad\\r\\nBad\\r\\nGreat\\r\\nGreat\\r\\n'
        """
        if self.state == ResultsStates.SELECTED:
            header, rows = self._make_tabular(remove_prefix)

            if filename is not None:
                with open(filename, "w") as f:
                    writer = csv.writer(f)
                    writer.writerow(header)
                    writer.writerows(rows)
            else:
                output = StringIO()
                writer = csv.writer(output)
                writer.writerow(header)
                writer.writerows(rows)

                if download_link:
                    import base64
                    from IPython.display import HTML, display

                    csv_file = output.getvalue()
                    b64 = base64.b64encode(csv_file.encode()).decode()
                    download_link = f'<a href="data:file/csv;base64,{b64}" download="my_data.csv">Download CSV file</a>'
                    display(HTML(download_link))
                else:
                    return output.getvalue()
        else:
            return self.select().to_csv(
                filename, remove_prefix=remove_prefix, download_link=download_link
            )

    def to_pandas(self, remove_prefix=False):
        csv_string = self.to_csv(remove_prefix=remove_prefix)
        csv_buffer = io.StringIO(csv_string)
        df = pd.read_csv(csv_buffer)
        return df

    def to_dicts(self, remove_prefix=False):
        df = self.to_pandas(remove_prefix=remove_prefix)
        df = df.convert_dtypes()
        list_of_dicts = df.to_dict(orient="records")
        # Convert any pd.NA values to None
        list_of_dicts = [
            {k: (None if pd.isna(v) else v) for k, v in record.items()}
            for record in list_of_dicts
        ]
        return list_of_dicts

    def tolist(self):
        print("Switch to using to_list() instead of tolist()")
        self.to_list()

    def to_list(self):
        # new_data = list(self[0].values())[0]
        if len(self) == 1:
            return list(self[0].values())[0]
        else:
            return tuple([list(x.values())[0] for x in self])

    def to_pivot(self):
        from pivottablejs import pivot_ui

        df = self.to_pandas()
        return pivot_ui(df)



from __future__ import annotations
import copy
import json
from typing import Any, Callable, Optional, Union
from edsl.exceptions import (
    AgentAttributeLookupCallbackError,
    AgentCombinationError,
    AgentRespondedWithBadJSONError,
)
from edsl.language_models import LanguageModel, LanguageModelOpenAIThreeFiveTurbo
from edsl.scenarios import Scenario
from edsl.utilities import (
    create_valid_var_name,
    dict_to_html,
    print_dict_as_html_table,
    print_dict_with_rich,
)


DEFAULT_INSTRUCTION = (
    """You are answering questions as if you were a human. Do not break character."""
)


class Agent:
    """An agent answers."""

    def __init__(
        self,
        traits: dict = None,
        attribute_lookup_callback: Callable = None,
        verbose: bool = False,
        original_names_dict: dict = None,
        instruction: str = None,
    ):
        self.attribute_lookup_callback = attribute_lookup_callback
        self.instruction = instruction or DEFAULT_INSTRUCTION
        self.verbose = verbose
        self._original_names = original_names_dict or dict()
        self._questions_to_traits = dict()
        self._traits = traits or dict()

    ################
    # TRAITS METHODS
    ################
    @property
    def traits(self) -> dict[str, Any]:
        """Returns the traits of the agent."""
        return self._traits

    def update_traits(
        self,
        new_attributes: list[str],
        attribute_lookup_callback: Callable = None,
    ) -> None:
        """Updates the Agent traits with new attributes. Allows an optional callback, defaulting to the instance's attribute_lookup_callback."""
        attribute_lookup = attribute_lookup_callback or self.attribute_lookup_callback
        if not attribute_lookup:
            raise AgentAttributeLookupCallbackError("Missing callback.")

        _traits = {}
        for attribute in new_attributes:
            try:
                trait_name, value = attribute_lookup(attribute)
                _traits[trait_name] = value
            except Exception as e:
                raise AgentAttributeLookupCallbackError(
                    f"Error in attribute lookup callback: {e}"
                    f"Attribute Lookup Callback: {attribute_lookup}"
                    f"Attribute: {attribute}"
                )
        self._traits = _traits

    def get_original_name(self, trait: str) -> str:
        """Returns the original name of a trait, if any."""
        return self._original_names.get(trait, trait)

    def agent_with_valid_trait_names(self) -> Agent:
        """
        Returns a new agent with trait keys that are valid variable names.
        Useful down the road; stores the original names.
        """
        original_names_dict = {}
        new_traits_dict = {}
        for key, value in self._traits.items():
            new_name = create_valid_var_name(key)
            new_traits_dict[new_name] = value
            original_names_dict[new_name] = key

        return Agent(traits=new_traits_dict, original_names_dict=original_names_dict)

    ################
    # LLM METHODS
    ################
    def construct_system_prompt(self) -> str:
        """Constructs the system prompt for the LLM call."""
        instruction = self.instruction
        traits = f"Your traits are: {self.traits}."
        return f"{instruction} {traits}"

    def get_response(
        self, prompt: str, system_prompt: str, model: LanguageModel = None
    ):
        """Calls the LLM and gets a response. Used in the `answer_question` method."""
        model = model or LanguageModelOpenAIThreeFiveTurbo(use_cache=True)

        try:
            response = model.get_response(prompt, system_prompt)
        except json.JSONDecodeError as e:
            raise AgentRespondedWithBadJSONError(
                f"Returned bad JSON: {e}"
                f"Prompt: {prompt}"
                f"System Prompt: {system_prompt}"
            )

        return response

    def answer_question(
        self,
        question: Question,
        scenario: Optional[Scenario] = None,
        model: Optional[LanguageModel] = None,
        debug: bool = False,
    ):
        """
        The main function that the agent calls to answer a question.
        1) It first constructs the prompts, then calls the LLM.
        2) It then validates the response and the answers.
        The answer from the JSON can be in code, so it uses the
        question-specific translation function to translate the code to the answer.
        """
        # simulated answers (w/o API call)
        #  if debug mode
        if debug:
            return question.simulate_answer(human_readable=True)
        #  if the question has this method
        if hasattr(question, "answer_question_directly"):
            return question.answer_question_directly(
                scenario=scenario, agent_traits=self.traits
            )
        #  if the agent has this method
        if hasattr(self, "answer_question_directly"):
            answer = self.answer_question_directly(question.question_name)
            response = {"answer": answer}
            response = question.validate_response(response)
            response["model"] = "human"
            response["scenario"] = scenario
            return response

        # actual answers (w/ API call)
        #  get answer
        scenario = scenario or Scenario()
        system_prompt = self.construct_system_prompt()
        prompt = question.get_prompt(scenario)
        response = self.get_response(prompt, system_prompt, model)
        #  validate answer
        response = question.validate_response(response)
        response = question.validate_answer(response)
        answer_code = response["answer"]
        response["answer"] = question.translate_answer_code_to_answer(
            answer_code, scenario
        )
        return response

    ################
    # Dunder Methods
    ################
    def __add__(self, other_agent: Agent = None) -> Agent:
        """
        Combines two agents by joining their traits
        - The agents must not have overlapping traits.
        """
        if other_agent is None:
            return self
        elif common_traits := set(self.traits.keys()) & set(other_agent.traits.keys()):
            raise AgentCombinationError(
                f"The agents have overlapping traits: {common_traits}."
            )
        else:
            new_agent = Agent(traits=copy.deepcopy(self.traits))
            new_agent.traits.update(other_agent.traits)
            return new_agent

    def __eq__(self, other: Agent) -> bool:
        """Checks if two agents are equal. Only checks the traits."""
        return self.traits == other.traits

    def __repr__(self) -> str:
        """Returns the agent's representation."""
        return f"Agent(traits = {self.traits})"

    ################
    # Forward methods
    ################
    def to(self, question_or_survey) -> Jobs:
        """Sends an agent to a question or survey."""
        from edsl.agents.AgentList import AgentList

        a = AgentList([self])
        return a.to(question_or_survey)

    def get_value(self, jobs: Jobs) -> list[Agent]:
        """Get a list of agents from a Jobs object. Used in Jobs.by()"""
        return jobs.agents

    def set_value(self, jobs: Jobs, new_values: list[Agent]) -> None:
        """Set the Jobs.agents attribute to the new values. Used in Jobs.by()"""
        jobs.agents = new_values

    ################
    # SERIALIZATION METHODS
    ################
    def to_dict(self) -> dict[str, Union[dict, bool]]:
        """Serializes to a dictionary."""
        return {"traits": self.traits, "verbose": self.verbose}

    @classmethod
    def from_dict(cls, agent_dict: dict[str, Union[dict, bool]]) -> Agent:
        """Deserializes from a dictionary."""
        return cls(**agent_dict)

    ################
    # DISPLAY Methods
    ################
    def dict_to_html(self) -> str:
        """Returns the agent's traits as an HTML table."""
        return dict_to_html(self.traits)

    def print(self, html: bool = False, show: bool = False) -> Optional[str]:
        """Prints the agent's traits as a table."""
        if html:
            return print_dict_as_html_table(self.traits, show)
        else:
            print_dict_with_rich(self.traits)


def main():
    """Consumes API credits"""
    from edsl.agents import Agent
    from edsl.questions import QuestionMultipleChoice

    # a simple agent
    agent = Agent(traits={"age": 10, "hair": "brown", "height": 5.5})
    agent.traits
    agent.print()
    # combining two agents
    agent = Agent(traits={"age": 10}) + Agent(traits={"height": 5.5})
    agent.traits
    # Agent -> Job using the to() method
    agent = Agent(traits={"allergies": "peanut"})
    question = QuestionMultipleChoice(
        question_text="Would you enjoy a PB&J?",
        question_options=["Yes", "No"],
        question_name="food_preference",
    )
    job = agent.to(question)
    # run the job
    results = job.run()
    # results



import random
import textwrap
from pydantic import BaseModel, Field, field_validator, model_validator
from typing import Type, Union
from edsl.questions.Question import Question
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings
from edsl.utilities.utilities import random_string


class QuestionMultipleChoiceGrid(Question):
    """This is a question where the answer is a matrix.
    It looks like this:

            | Yuk   | Yum   |
    --------+-------+-------+
    Hot Dog |   X   |       |
    --------+-------+-------+
    Pizza   |       |   X   |
    --------+-------+-------+

    There are 2 types of matrix questions (Google Forms as example):
    - Checkbox grid: any number of column selections per row allowed
    - Multiple choice grid: only 1 column selection per row allowed

    You have to select one option from each row and column if answer required.
    Otherwise, you can leave a row blank.
    """

    question_type = "multiple_choice_grid"

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked the following multiple choice question in the form of a matrix: 
        {{question_text}}
        The matrix rows are 
        {% for option in row_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}                       
        The matrix columns are 
        {% for option in column_options %}
        {{ loop.index0 }}: {{option}}
        {% endfor %}   
        {% if required == True %}
        You are required to provide an answer.
        {% endif %}     
        Return a valid JSON formatted like this using the numbers of the options: 
        {"answer": <[put list of each column option selected for each row here]>, 
        "comment": "<put explanation here>"}
        Example answer: [1,2]
        """
        )

    def translate_answer_code_to_answer(self, answer_codes):
        """Translates the answer codes to the actual answers.
        For example, for a budget question with options ["a", "b", "c"],
        the answer codes are 0, 1, and 2. The LLM will respond with 0.
        This code will translate that to "a".
        """
        translated_codes = []
        for answer_code_list in answer_codes:
            translated_answer_code_list = []
            for answer_code in answer_code_list:
                translated_answer_code_list.append(
                    self.column_options[int(answer_code)]
                )
            translated_codes.append(translated_answer_code_list)

        return translated_codes

    @classmethod
    def construct_question_data_model(cls) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionMultipleChoiceGrid"""

            question_text: str = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            row_options: list[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_NUM_OPTIONS
            )
            column_options: list[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_NUM_OPTIONS
            )
            required: bool = False

            @field_validator("row_options")
            def check_unique(cls, value):
                return cls.base_validator_check_unique(value)

            @field_validator("row_options")
            def check_option_string_lengths(cls, value):
                return cls.base_validator_check_option_string_lengths(value)

            @field_validator("column_options")
            def check_unique(cls, value):
                return cls.base_validator_check_unique(value)

            @field_validator("column_options")
            def check_option_string_lengths(cls, value):
                return cls.base_validator_check_option_string_lengths(value)

        return LocalQuestionData

    def construct_answer_data_model(self) -> Type[BaseModel]:
        row_options = self.row_options
        column_options = self.column_options
        required = self.required

        class LocalAnswerDataModel(AnswerData):
            answer: list[Union[int, None]]

            @model_validator(mode="after")
            def check_required_answer(self):
                if required == True and len(self.answer) < len(row_options):
                    raise ValueError("There are missing answers.")
                return self

            @model_validator(mode="after")
            def check_answer_count(self):
                if len(self.answer) > len(row_options):
                    raise ValueError("There are too many answers.")
                return self

        return LocalAnswerDataModel

    def simulate_answer(self):
        # Expand to allow multiple selections for checkbox (non-mc) case
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        answers = []
        for _ in self.row_options:
            answers.append([random.choice(range(len(self.column_options)))])
        return {"answer": answers, "comment": random_string()}


if __name__ == "__main__":
    from edsl.exceptions import (
        QuestionResponseValidationError,
        QuestionAnswerValidationError,
    )

    print("Now validating multiple choice grid questions")

    q = QuestionMultipleChoiceGrid(
        question_text="Indicate how much you like the following foods.",
        row_options=["Pizza", "Ice Cream"],
        column_options=["Very much", "Somewhat", "Not at all", "Don't know"],
        required=True,
    )
    print(q.get_prompt())
    response = {"answer": [0, 1], "comment": "OK"}
    print(response)
    q.validate_response(response)
    response.pop("comment")
    q.validate_answer(response)
    print("This is a valid response.\n")

    print("Now checking a non-required question.")
    q = QuestionMultipleChoiceGrid(
        question_text="Indicate how much you like the following foods.",
        row_options=["Pizza", "Ice Cream"],
        column_options=["Very much", "Somewhat", "Not at all", "Don't know"],
        required=False,
    )
    print(q.get_prompt())
    response = {"answer": [None, 2], "comment": "OK"}
    print(response)
    q.validate_response(response)
    response.pop("comment")
    q.validate_answer(response)
    print("This is a valid response.\n")

    print("Now checking a required question.")
    q = QuestionMultipleChoiceGrid(
        question_text="Indicate how much you like the following foods.",
        row_options=["Pizza", "Ice Cream"],
        column_options=["Very much", "Somewhat", "Not at all", "Don't know"],
        required=True,
    )
    print(q.get_prompt())
    try:
        response = {"answer": [3, None], "comment": "OK"}
        print(response)
        q.validate_response(response)
        response.pop("comment")
        q.validate_answer(response)
    except QuestionAnswerValidationError:
        print("Caught bad answer.\n")

    print(q.get_prompt())
    try:
        response = {"answer": [["Mon"], ["Wed"]], "comment": "OK"}
        print(response)
        q.validate_response(response)
        response.pop("comment")
        q.validate_answer(response)
    except QuestionAnswerValidationError:
        print("Caught bad answer.\n")



import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Optional, Type
from edsl.exceptions import QuestionAnswerValidationError
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.utilities.utilities import random_string


class QuestionFreeText(QuestionData):
    """Pydantic data model for QuestionFreeText"""

    allow_nonresponse: Optional[bool] = False

    def __new__(cls, *args, **kwargs) -> "QuestionFreeTextEnhanced":
        # runs before __init__
        #   create an instance of QuestionFreeText using the __new__ method from QuestionData
        instance = super(QuestionFreeText, cls).__new__(cls)
        #   call QuestionFreeText's __init__
        #   (Note: Python will not __init__ again after __new__)
        instance.__init__(*args, **kwargs)
        #   pass the validated QuestionFreeText to create a QuestionFreeTextEnhanced
        return QuestionFreeTextEnhanced(instance)

    def __init__(self, **data):
        # call QuestionData's __init__ to run its validators
        super().__init__(**data)


class QuestionFreeTextEnhanced(Question):
    question_type = "free_text"

    def __init__(self, question: QuestionFreeText):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        Return a valid JSON formatted like this: 
        {"answer": "<put free text answer here>"}
        """
        )

    def translate_answer_code_to_answer(self, answer, scenario):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class QuestionFreeTextAnswerDataModel(AnswerData):
            answer: str = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise QuestionAnswerValidationError(
                        "Empty answer to a `QuestionFreeText` question, but it was not allowed."
                    )
                return value

        return QuestionFreeTextAnswerDataModel

    ################
    # Less important
    ################

    def simulate_answer(self) -> dict[str, str]:
        return {"answer": random_string()}

    def form_elements(self) -> str:
        html_output = f"""
        <label for="{self.question_name}">{self.question_text}</label>
        <div id="{self.question_name}_div">
            <input type="text" id="{self.question_name}" name="{self.question_text}">
        </div>
        """
        return html_output


# if __name__ == 'main':
#     from edsl.questions import QuestionFreeText
#     q = QuestionFreeText(question_text = "How old are you?", question_name = "age")
#     result = q.run()
#     print(result)



import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Type, Optional
from edsl.questions.Question import Question
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings
from edsl.utilities.utilities import random_string


class QuestionShortAnswer(QuestionData):
    """Pydantic data model for QuestionShortAnswer"""

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )
    expression_constraint: Optional[str] = Field(
        ..., min_length=0, max_length=Settings.MAX_EXPRESSION_CONSTRAINT_LENGTH
    )
    allow_nonresponse: bool = None

    def __new__(cls, *args, **kwargs):
        # this runs before __init__
        # we create an instance of the Pydantic data model
        instance = super(QuestionShortAnswer, cls).__new__(cls)
        # we run it through __init__, which will run the validators
        instance.__init__(*args, **kwargs)
        # we pass the validated the instance to create a QuestionMultipleChoiceEnhanced
        # which has all the abstract base clas methods we need
        return QuestionShortAnswerEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)


class QuestionShortAnswerEnhanced(Question):
    question_type = "short_answer"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked the following question: {{question_text}}
        {% if expression_constraint %}
        Your response must satisfy the following constraints:
        {{expression_constraint}}
        {% endif %}
        Return a valid JSON formatted exactly like this: 
        {"answer": "<put your answer here>", "comment": "<put explanation here>"}         
        """
        )

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: str = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise ValueError("You must provide a response.")
                return value

        return LocalAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        return {"answer": random_string()}

    def form_elements(self):
        raise NotImplementedError


if __name__ == "__main__":
    q = QuestionShortAnswer(
        question_text="What do you want to do today?",
        expression_constraint=None,
        question_name="today",
    )
    results = q.run()
    print(results)



import re
import textwrap
from random import randint
from pydantic import BaseModel, field_validator
from typing import Type, Union
from edsl.questions import Question, QuestionData, AnswerData
from edsl.exceptions import QuestionAnswerValidationError
from edsl.utilities.utilities import random_string


from edsl.utilities.utilities import random_string


class QuestionNumerical(QuestionData):
    """Pydantic data model for QuestionNumerical"""

    min_value: Union[int, float, None] = None
    max_value: Union[int, float, None] = None

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionNumericalEnhanced":
        instance = super(QuestionNumerical, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionNumericalEnhanced(instance)


class QuestionNumericalEnhanced(Question):
    question_type = "numerical"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        # make the min/max conditional parts of the instructions {% if ... end if}
        return textwrap.dedent(
            """\
        You are being asked a question that requires a numerical response 
        in the form of an integer or decimal (e.g., -12, 0, 1, 2, 3.45, ...).
        
        Your response must be in the following format:
        {"answer": "<your numerical answer here>", "comment": "<your explanation here"}

        You must only include an integer or decimal in the quoted "answer" part of your response. 

        Here is an example of a valid response:
        {"answer": "100", "comment": "This is my explanation..."}

        Here is an example of a response that is invalid because the "answer" includes words:
        {"answer": "I don't know.", "comment": "This is my explanation..."}

        If your response is equivalent to zero, your formatted response should look like this:
        {"answer": "0", "comment": "This is my explanation..."}
        
        You are being asked the following question: {{question_text}}
        {% if min_value is not none %}
        Minimum answer value: {{min_value}}
        {% endif %}
        {% if max_value is not none %}
        Maximum answer value: {{max_value}}
        {% endif %}
        """
        )
        """
        Simpler default prompt that works with GPT-4:

        You are being asked the following question: {{question_text}}
        {% if min_value is not none %}
        Minimum answer value: {{min_value}}
        {% endif %}
        {% if max_value is not none %}
        Maximum answer value: {{max_value}}
        {% endif %}
        Return a valid JSON formatted like this: 
        {"answer": "<put integer or float answer here>", "comment": "<put explanation here>"}
        """

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class QuestionNumericalAnswerDataModel(AnswerData):
            answer: Union[int, float, str]

            @field_validator("answer")
            def check_answer_numeric(cls, value):
                if type(value) == str:
                    value = value.replace(",", "")
                    value = "".join(re.findall(r"[-+]?\d*\.\d+|\d+", value))
                    if value.isdigit():
                        value = int(value)
                    else:
                        try:
                            float(value)
                            value = float(value)
                        except ValueError:
                            raise QuestionAnswerValidationError(
                                f"Answer should be numerical (int or float)."
                            )
                return value

            @field_validator("answer")
            def check_answer(cls, value):
                # breakpoint()
                if self.min_value is not None and value < self.min_value:
                    raise QuestionAnswerValidationError(
                        f"Value {value} is less than {self.min_value}"
                    )
                if self.max_value is not None and value > self.max_value:
                    raise QuestionAnswerValidationError(
                        f"Value {value} is greater than {self.max_value}"
                    )
                return value

        return QuestionNumericalAnswerDataModel

    ################
    # Less important
    ################

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def simulate_answer(self, human_readable=True):
        "Simulates a valid answer for debugging purposes"
        return {"answer": randint(0, 100), "comment": random_string()}

    def form_elements(self):
        html_output = f"\n\n\n<label>{self.question_text}</label>\n"
        html_output += f"""<div id="{self.question_name}_div">
        <input type="number" id="{self.question_name}" name="{self.question_text}">
        </div>\n"""
        return html_output


if __name__ == "__main__":
    from edsl.agents import Agent

    sadness_scale = QuestionNumerical(
        question_name="sadness_scale",
        question_text="On a scale of 1-10 how often do you feel sad or down ?",
        min_value=0,
        max_value=10,
    )
    results = sadness_scale.by(Agent(traits={"sadness": "Very, very low"})).run()
    print(results)



import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Optional, Type, Union
from edsl.questions import Question, QuestionData, AnswerData, Settings
from edsl.exceptions import QuestionAnswerValidationError
from edsl.utilities.utilities import random_string


class QuestionList(QuestionData):
    """Pydantic data model for QuestionList"""

    allow_nonresponse: Optional[bool] = False
    max_list_items: Optional[int] = None

    # see QuestionFreeText for an explanation of how __new__ works
    def __new__(cls, *args, **kwargs) -> "QuestionListEnhanced":
        instance = super(QuestionList, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionListEnhanced(instance)

    def __init__(self, **data):
        super().__init__(**data)


class QuestionListEnhanced(Question):
    question_type = "list"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    @property
    def instructions(self) -> str:
        return textwrap.dedent(
            """\
        {{question_text}}

        Your response should be only a valid JSON of the following format:
        {
            "answer": <list of comma-separated words or phrases >, 
            "comment": "<put comment here>"
        }
        {% if max_list_items is not none %}
        The list must not contain more than {{ max_list_items }} items.
        {% endif %}
        """
        )

    def translate_answer_code_to_answer(self, answer, scenario):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        "Constructs the answer data model for this question"

        class QuestionListAnswerDataModel(AnswerData):
            answer: Union[list[str], list[dict]] = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer_nonresponse(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == [] or value is None)
                ):
                    raise QuestionAnswerValidationError("You must provide a response.")

                return value

            @field_validator("answer")
            def check_answer_length(cls, value):
                if (
                    hasattr(self, "max_list_items")
                    and self.max_list_items is not None
                    and (len(value) > self.max_list_items)
                ):
                    raise QuestionAnswerValidationError("Response has too many items.")
                return value

            @field_validator("answer")
            def check_answer_check(cls, value):
                if any([item == "" for item in value]):
                    raise QuestionAnswerValidationError(
                        f"Answer cannot contain empty strings, but got {value}."
                    )
                return value

        return QuestionListAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        return {"answer": [random_string(), random_string()]}

    def form_elements(self):
        html_output = f"""
        <label for="{self.question_name}">{self.question_text}</label>
        <div id="{self.question_name}_div">
            <input type="text" id="{self.question_name}" name="{self.question_text}">
        </div>
        """
        return html_output


def main():  # pragma: no cover
    """This is a demo of how to use the QuestionList. It consumes API calls."""
    from edsl.language_models import LanguageModelOpenAIFour
    from edsl.questions import QuestionList

    m4 = LanguageModelOpenAIFour()

    question = QuestionList(
        question_text="What are some factors that could determine whether someone likes ice cream?",
        question_name="ice_cream_love_attributes",
        max_list_items=20,
    )

    results = question.by(m4).run()
    results.data[0].answer

    question = QuestionList(
        question_text="What are the most beloved sport teams in America?",
        question_name="most_beloved_teams",
        max_list_items=5,
    )

    results = question.by(m4).run()
    results.data[0].answer



from __future__ import annotations
import re
import textwrap
import uuid
from abc import ABC, abstractmethod
from jinja2 import Template, Environment, meta
from pydantic import BaseModel, ValidationError
from typing import Any, Type, Union
from edsl.exceptions import (
    QuestionAnswerValidationError,
    QuestionAttributeMissing,
    QuestionResponseValidationError,
    QuestionSerializationError,
    QuestionScenarioRenderError,
)
from edsl.questions.question_registry import get_question_class
from edsl.questions.utils import LLMResponse
from edsl.utilities.utilities import HTMLSnippet


class Question(ABC):
    """
    Methods and requirements for children-Questions.

    Children must implement constuctors for two Pydantic data models:
    - `QuestionData` used to validate the data that creates a question
    - `AnswerData` used to validate the answers, and can depend on `QuestionData`

    Some notes on the attributes
    - `question_type` registers the question type.
    - `instructions` are directed towards agents to help construct prompts.

    There is a validation step for the LLM response:
    - The LLM should always return valid JSON with two fields: "answer" and "comments"
    - Use indices instead of the actual text of the option keys. The reason to use indices instead of
      the actual text is that it will make type checking of LLM responses easier.
    - The "answer" field should be valid for the question type.

    Question responses are required by default but can be specified as required=False at construction.
    """

    def __init__(self, question: BaseModel):
        # create class attributes for all fields of the QuestionData pydantic model
        for key, value in question.model_dump().items():
            setattr(self, key, value)
        # store the dict representation of the QuestionData pydantic model
        self.data = question.model_dump()
        # use QuestionData to construct the AnswerData pydantic model
        self.answer_data_model = self.construct_answer_data_model()
        self.uuid = str(uuid.uuid4())
        self.sanity_checks()

    def to_dict(self) -> dict:
        """Converts a Question to a dictionary"""
        # TODO: This is a very specific dictionary, which should be a model
        # TODO: having {type:str, data:dict}
        data = self.data
        data["type"] = self.question_type
        return data

    @classmethod
    def from_dict(cls, data: dict) -> Question:
        """Constructs a Question from the dictionary created by the `to_dict` method"""
        # do not alter the original data
        local_data = data.copy()
        try:
            question_type = local_data.pop("type")
        except:
            raise QuestionSerializationError(
                "Question data does not have a 'type' field"
            )
        # Use the get_question_class function to lazily load the class
        question_class = get_question_class(question_type)
        return question_class(**local_data)

    def __repr__(self):
        class_name = self.__class__.__name__.replace("Enhanced", "")
        items = [
            f"{k} = '{v}'" if isinstance(v, str) else f"{k} = {v}"
            for k, v in self.data.items()
            if k != "type"
        ]
        return f"{class_name}({', '.join(items)})"

    # TODO: Throws an error that should be addressed at QuestionFunctional
    def __add__(self, other_question):
        """
        Composes two questions into a single question.
        >>> from edsl.scenarios.Scenario import Scenario
        >>> from edsl.questions.QuestionFreeText import QuestionFreeText
        >>> from edsl.questions.QuestionNumerical import QuestionNumerical
        >>> q1 = QuestionFreeText(question_text = "What is the capital of {{country}}", question_name = "capital")
        >>> q2 = QuestionNumerical(question_text = "What is the population of {{capital}}, in millions. Please round", question_name = "population")
        >>> q3 = q1 + q2
        >>> Scenario({"country": "France"}).to(q3).run().select("capital_population")
        ['2']
        """
        from edsl.questions import compose_questions

        return compose_questions(self, other_question)

    @property
    @abstractmethod
    def instructions(self) -> str:  # pragma: no cover
        """
        Instructions for each question.
        - the values are question type-specific
        - the templating standard is Jinja2.
        - it is necessary to include both "answer" and "comment" as the only keys.
        - Note: children should implement this method as a property.

        An example for `QuestionFreeText`:
         You are being asked the following question: {{question_text}}
         Return a valid JSON formatted like this:
         {"answer": "<your free text answer>", "comment": "<put explanation here>"}
        """
        pass

    @staticmethod
    def scenario_render(text: str, scenario_dict: dict) -> str:
        """
        Scenarios come in as dictionaries. This function goes through the text of a question
        we are presenting to the LLM and replaces the variables with the values from the scenario.
        Because we allow for nesting, we need to do this many times.
        - We hard-code in a nesting limit of 100 just because if we hit that, it's probably a bug and not that some lunatic has actually made a 100-deep nested question.
        """
        t = text
        MAX_NESTING = 100
        counter = 0
        while True:
            counter += 1
            new_t = Template(t).render(scenario_dict)
            if new_t == t:
                break
            t = new_t
            if counter > MAX_NESTING:
                raise QuestionScenarioRenderError(
                    "Too much nesting - you created an infnite loop here, pal"
                )

        return new_t

    def get_prompt(self, scenario=None) -> str:
        """Shows which prompt should be used with the LLM for this question.
        It extracts the question attributes from the instantiated question data model.
        """
        if scenario is None:
            scenario = {}

        template = Template(self.instructions)
        attributes = self.__dict__
        template_with_attributes = template.render(attributes)

        env = Environment()
        ast = env.parse(template_with_attributes)
        undeclared_variables = meta.find_undeclared_variables(ast)
        if any([v not in scenario for v in undeclared_variables]):
            raise QuestionScenarioRenderError(
                f"Scenario is missing variables: {undeclared_variables}"
            )

        prompt = self.scenario_render(template_with_attributes, scenario)
        return prompt

    @abstractmethod
    def translate_answer_code_to_answer(self):  # pragma: no cover
        """Translates the answer code to the actual answer. Behavior depends on the question type."""
        pass

    @abstractmethod
    def construct_answer_data_model(self) -> Type[BaseModel]:  # pragma: no cover
        """Constructs the Pydantic data model for the answer"""
        pass

    def validate_response(self, response: dict) -> Union[dict, None]:
        """Validates the LLM response."""
        try:
            _ = LLMResponse(**response)
        except Exception as e:
            print(e)
            print(f"Offending response: {response}")
            raise QuestionResponseValidationError("Invalid response")

        return response

    def validate_answer(self, answer: Any) -> Union[dict, None]:
        """Validates the answer to the question using the constructed answer_data_model"""
        try:
            return self.answer_data_model(**answer).model_dump()
        except ValidationError as e:
            raise QuestionAnswerValidationError(f"Invalid answer {answer}/nError: {e}")

    @abstractmethod
    def simulate_answer(self, human_readable=True) -> dict:  # pragma: no cover
        """Simulates a valid answer for debugging purposes (what the validator expects)"""
        pass

    def formulate_prompt(self, traits=None, focal_item=None):
        """
        Builds the prompt to send to the LLM. The system prompt contains:
        - Context that might be helpful
        - The traits of the agent
        - The focal item and a description of what it is.
        """
        system_prompt = ""

        instruction_part = textwrap.dedent(
            """\
        You are answering questions as if you were a human. 
        Do not break character.  
        """
        )
        system_prompt += instruction_part

        if traits is not None:
            relevant_trait = traits.relevant_traits(self)
            traits_part = f"Your traits are: {relevant_trait}"
            system_prompt += traits_part

        prompt = ""
        if focal_item is not None:
            focal_item_prompt_fragment = textwrap.dedent(
                f"""\
            The question you will be asked will be about a {focal_item.meta_description}.
            The particular one you are responding to is: {focal_item.content}.
            """
            )
            prompt += focal_item_prompt_fragment

        prompt += self.get_prompt()

        return prompt, system_prompt

    def sanity_checks(self) -> None:
        """Some sanity checks for a Question."""
        SPECIAL = ["QuestionFunctionalEnhanced"]
        expected_attributes = ["question_name", "question_text"]
        if self.__class__.__name__ not in SPECIAL:
            # Check if all expected attributes are present
            for attribute in expected_attributes:
                if attribute not in self.__dict__:
                    raise QuestionAttributeMissing(
                        f"Question {self.__class__} is missing attribute {attribute}"
                    )
            # Check for unescaped curly braces in `question_text`
            matches_single_but_not_double = r"(?<!\\)(?<!{){(?!{)|(?<!\\)(?!})}(?!})"
            if bool(re.search(matches_single_but_not_double, self.question_text)):
                print("WARNING: unescaped curly braces in `question_text`.")
                print("You probably meant to use **double** curly braces.")

    @abstractmethod
    def form_elements(self):  # pragma: no cover
        """Returns the HTML that helps to put this question in a form"""
        pass

    ################
    # Question -> Survey methods
    ################
    def add_question(self, other):
        "Adds a question to this question by turning them into a survey with two questions"
        from edsl.surveys.Survey import Survey

        s = Survey([self, other], [self.question_name, other.question_name])
        return s

    def run(self, *args, **kwargs):
        "Turns a single question into a survey and run it."
        from edsl.surveys.Survey import Survey

        s = Survey([self], [self.question_name])
        return s.run(*args, **kwargs)

    def by(self, *args):
        "This turns a single question into a survey and runs it."
        from edsl.surveys.Survey import Survey

        s = Survey([self], [self.question_name])
        return s.by(*args)

    ################
    # Less important
    ################
    def html(self, submit_url=None):
        submit_button = f'<input type="submit" value="Submit">\n' if submit_url else ""
        html_string = (
            f'<form id="{self.question_name}" action="{submit_url if submit_url else ""}" method="post">'
            f"<div>{self.form_elements()}</div>"
            f'<input type="hidden" name="question_name" value="{self.question_name}">'
            f"{submit_button}"
            "</form>"
        )
        return HTMLSnippet(html_string)



import pandas as pd
import statsmodels.api as sm
import re
import numpy as np


def extract_variable_names(formula):
    # Use regular expressions to find all variable names within the formula
    variable_names = re.findall(r"\b\w+\b", formula)
    # what do we do if there is a variable named 'C'?
    return [variable_name for variable_name in variable_names if variable_name != "C"]


class RegressionMixin:
    def regression(self, formula):
        """
        This runs a linear regression
        https://www.statsmodels.org/stable/index.html
        formula = 'y ~ x1 + x2 + x3 + C(x4)'
        """

        variable_names = extract_variable_names(formula)

        df = self.select(*variable_names).to_pandas()

        # get rid of the headers
        column_mapping = {col: col.split(".")[-1] for col in df.columns}
        # Rename the columns using the dictionary
        df.rename(columns=column_mapping, inplace=True)

        model = sm.OLS.from_formula(formula, data=df)
        regression_result = model.fit()

        return regression_result


if __name__ == "__main__":
    formula = "y ~ x1 + x2 + C(customer_service)"
    print(extract_variable_names_with_indicator(formula))



from __future__ import annotations
from collections import UserList
from docx import Document
from typing import Optional, Union
from edsl.agents import Agent


class AgentList(UserList):
    def __init__(self, data: Optional[list] = None):
        if data is not None:
            super().__init__(data)
        else:
            super().__init__()

    def to(self, question_or_survey: Union["Question", "Survey"]):
        return question_or_survey.by(*self)

    def update_traits(self, new_attributes: list):
        for agent in self.data:
            agent.update_traits(new_attributes)

    def print(self, html=False):
        html = ""
        for agent in self.data:
            html += agent.dict_to_html()
        from edsl.utilities.interface import gen_html_sandwich
        from edsl.utilities.interface import view_html

        view_html(gen_html_sandwich(html))

    def to_dict(self):
        return {"agent_list": [agent.to_dict() for agent in self.data]}

    @classmethod
    def from_dict(cls, data: dict) -> "AgentList":
        """Deserializes the dictionary back to an Agent List object."""
        agents = [Agent.from_dict(agent_dict) for agent_dict in data["agent_list"]]
        return cls(agents)

    def docx(self) -> Document:
        "Generates a docx document for the survey"
        doc = Document()

        doc.add_heading("EDSL Auto-Generated Agent Description")

        doc.add_paragraph(f"\n")

        for index, agent in enumerate(self.data):
            # Add question as a paragraph
            h = doc.add_paragraph()
            h.add_run(f"Agent {index + 1}").bold = True

            p = doc.add_paragraph()
            # p.add_run(agent.persona)
            for key, value in agent.traits.items():
                p.add_run(f"{key}: ").bold = True
                p.add_run(f"{value}\n")

        return doc


if __name__ == "__main__":
    a = AgentList([1, 2, 3])
    print(a)

    from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoice
    from edsl.agents.Agent import Agent

    q = QuestionMultipleChoice(
        question_text="How are you feeling?",
        question_options=["Very sad.", "Sad.", "Neutral.", "Happy.", "Very happy."],
        question_name="feelings",
    )

    a = Agent(traits={"feeling": "Very sad."})
    agent_list = AgentList([a])
    results = agent_list.to(q).run()

    print(results)

    results2 = q.by(agent_list).run()

    print(results2)



import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Optional, Type, Union
from edsl.questions.Question import Question
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings


class QuestionExtractDataLists(Question):
    """Extracts a specified data type from a list of texts, storing information
    extracted from each text as a new list, and then returning a list of those
    lists (which may be different lengths depending on the content of each text).
    Each list starts with the number of the text (starting from 0) and then the
    extracted information."""

    question_type = "extract_data_lists"

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked to extract certain information from a set of texts.
        {{question_text}}:
        {% for d in data %}
        {{ loop.index0 }}: {{d}}
        {% endfor %} 
        From each text, extract the following information as a new comma-separated list
        {{extract_data}} 
        Return all of the lists of extracted information in a valid JSON formatted exactly like this: 
        {"answer": [<put your list of lists here separated by commas>], "comment": "<put explanation here>"}         
        """
        )

    def __repr__(self):
        return f"""{self.__class__.__name__}(question_text = "{self.question_text}", extract_data = "{self.extract_data}", question_name = "{self.question_name}")"""

    @classmethod
    def construct_question_data_model(cls) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionExtractData"""

            question_text: Optional[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            extract_data: str = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            data: list[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            allow_nonresponse: bool = None

        return LocalQuestionData

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: list[list[Union[str, int]]] = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise ValueError("You must provide a response.")
                return value

        return LocalAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        raise NotImplementedError

    def form_elements(self):
        raise NotImplementedError

    # def __init__(self, question_text, extract_data, question_name=None):
    #     self.question_text = question_text
    #     self.extract_data = extract_data
    #     self.question_name = question_name

    #     d = {'question_text': question_text,
    #          'exptract_data': extract_data,
    #          'question_name': question_name}

    #     super().__init__(**d)


if __name__ == "__main__":
    from edsl import Agent, Scenario, print_dict_with_rich

    recipes = [
        """Classic Chocolate Chip Cookies
                Ingredients:

                1/2 cup (1 stick) unsalted butter, softened
                3/4 cup brown sugar, packed
                1/4 cup granulated sugar
                1 large egg
                2 teaspoons vanilla extract
                1.5 cups all-purpose flour
                1/2 teaspoon baking soda
                1/2 teaspoon salt
                1 cup chocolate chips
                1/2 cup chopped nuts (optional)
                Instructions:

                Preheat your oven to 350°F (175°C).
                In a large bowl, cream together the butter, brown sugar, and granulated sugar until light and fluffy.
                Beat in the egg and vanilla extract.
                In another bowl, whisk together the flour, baking soda, and salt. Gradually add this to the wet ingredients, stirring to incorporate.
                Fold in the chocolate chips and nuts, if using.
                Drop rounded tablespoons of dough onto a parchment-lined baking sheet, spacing them about 2 inches apart.
                Bake for 10-12 minutes or until golden around the edges but still soft in the center.
                Allow the cookies to cool on the baking sheet for a few minutes before transferring them to a wire rack to cool completely.""",
        """Easy Fruit Salad
                Ingredients:

                2 cups fresh pineapple chunks
                1 cup fresh strawberries, halved
                1 cup grapes, halved
                2 bananas, sliced
                1 apple, chopped
                1 cup yogurt or whipped cream for serving (optional)
                Instructions:

                In a large bowl, combine the pineapple, strawberries, grapes, bananas, and apple.
                Toss everything together until well mixed.
                Serve the fruit salad chilled, optionally topped with a dollop of yogurt or whipped cream.""",
        """Simple Rice Pudding
                Ingredients:

                1/2 cup uncooked white rice
                2 cups milk
                1/3 cup granulated sugar
                1/4 teaspoon salt
                1 egg, beaten
                2/3 cup raisins
                1 teaspoon vanilla extract
                1/2 teaspoon ground cinnamon (for serving)
                Instructions:

                In a medium saucepan, bring 1.5 cups of water to a boil. Stir in the rice and reduce the heat to low. Cover and simmer for 18-20 minutes, or until the rice is tender.
                Add the milk, sugar, and salt to the rice. Cook over medium heat until thick and creamy, 15 to 20 minutes, stirring frequently.
                Stir in the beaten egg and raisins. Cook for 2 more minutes, stirring constantly.
                Remove from heat and stir in the vanilla extract.
                Serve warm, sprinkled with a little ground cinnamon on top.""",
    ]

    q = QuestionExtractDataLists(
        question_text="The texts are recipes",
        extract_data="ingredients",
        data=recipes,
        question_name="ingredients",
    )
    scenarios = [Scenario({"recipe": recipe}) for recipe in recipes]
    print(q.get_prompt())
    results = q.by(*scenarios).run()
    print(results)

    # print(q.validate_answer(answer))
    # print(q.get_prompt())



from pydantic import BaseModel, Field, field_validator
from typing import Type
from edsl.exceptions import (
    QuestionCreationValidationError,
    QuestionAnswerValidationError,
)
from edsl.questions import Settings, QuestionData, AnswerData
from edsl.questions.QuestionMultipleChoice import QuestionMultipleChoiceEnhanced


class QuestionYesNo(QuestionData):
    """Pydantic data model for QuestionYesNo"""

    question_text: str = Field(
        ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
    )
    question_options: list = ["Yes", "No"]

    def __new__(cls, *args, **kwargs) -> "QuestionYesNoEnhanced":
        # see QuestionFreeText for an explanation of how __new__ works
        instance = super(QuestionYesNo, cls).__new__(cls)
        instance.__init__(*args, **kwargs)
        return QuestionYesNoEnhanced(instance)

    @field_validator("question_options")
    def check_successive(cls, value):
        if sorted(value) != ["No", "Yes"]:
            raise QuestionCreationValidationError(f"Options must be 'Yes' and 'No'")
        return value


class QuestionYesNoEnhanced(QuestionMultipleChoiceEnhanced):
    """Same as a QuestionCheckBox, but with question_options=["yes","no"]"""

    question_type = "yes_no"

    def __init__(self, question: BaseModel):
        super().__init__(question)

    def construct_answer_data_model(self) -> Type[BaseModel]:
        acceptable_values = list(range(len(self.question_options)))

        class QuestionYesNoAnswerDataModel(AnswerData):
            answer: int

            @field_validator("answer")
            def check_answer(cls, value):
                if value in acceptable_values:
                    return value
                else:
                    raise QuestionAnswerValidationError(
                        f"Answer {value} is not in the acceptable values {acceptable_values}"
                    )

        return QuestionYesNoAnswerDataModel



"""This question type is for option sets that are too large to read into the LLM.
We instead compute embeddings of the options, ask the LLM the question as free text, 
create an embedding of the LLM's answer and compare the embeddings, and ask 
the LLM to select the best option among the subset of close embeddings. 

Steps:
1. Compute embeddings for the option set.
2. Ask the LLM to answer the ultimate question (as free text).
3. Create an embedding of the LLM's answer.
4. Compare the embeddings.
5. Create a multiple choice question with the close embeddings as options.
6. Ask the LLM to answer the multiple choice question.
7. Validate the answer to the multiple choice question.

Can we turn a survey back into a question?
Can we question-ify a survey?

Use Agent differently
Similar to theme finder 


Consider
- Add a special attribute to a question
- Have agent check whether attribute exists
- Answer question the different way

Create a different kind of question that looks the same to the end user
but has a different way of getting answered.
"""



from collections import UserDict


class Result(UserDict):
    """Represents the result of an interview: one survey, one agent, one scenario,
    one model, one iteration, and one result."""

    def __init__(self, survey, agent, scenario, model, iteration, answer):
        # instantiates the UserDict with the actual objects
        data = {
            #            "survey": survey,
            "survey": None,
            "agent": agent.agent_with_valid_trait_names(),
            "scenario": scenario,
            "model": model,
            "iteration": iteration,
            "answer": answer,
        }

        super().__init__(**data)

        # Also assign the attributes to the class for convenience

        # TODO: This actually isn't expensive in terms of memory
        # for key, value in data.items():
        #    setattr(self, key, value)
        self.agent = agent

        ## TODO: Dictionary representations
        self.sub_dicts = {
            "answer": answer,
            "scenario": scenario,
            # "agent": agent.traits,
            "agent": agent.agent_with_valid_trait_names().traits,
            "model": model.parameters | {"model": model.model},
        }

        self.combined_dict = {}
        for key in self.sub_dicts:
            self.combined_dict.update(self.sub_dicts[key])
            self.combined_dict.update({key: self.sub_dicts[key]})

    def get_value(self, data_type, key):
        """Returns the value for a given key and data type e.g.,
        results.get_value("answer", "how_feeling") will return "Good" or "Bad" or whatnot
        """
        return self.sub_dicts[data_type][key]

    @property
    def key_to_data_type(self):
        d = {}
        for data_type in ["agent", "scenario", "model", "answer"]:
            for key in self.sub_dicts[data_type]:
                d[key] = data_type
        return d

    def __repr__(self):
        survey = self.survey if hasattr(self, "survey") else None
        return f"Result(survey={survey}, agent={self.agent}, scenario={self.scenario}, model={self.model}, iteration={self.iteration}, answer={self.answer})"

    def to_dict(self):
        return {
            k: v if not hasattr(v, "to_dict") else v.to_dict() for k, v in self.items()
        }

    @classmethod
    def from_dict(self, json_dict):
        from edsl.agents import Agent
        from edsl.scenarios import Scenario
        from edsl.language_models import LanguageModel

        return Result(
            # survey=Survey.from_dict(json_dict["survey"]),
            survey=None,
            agent=Agent.from_dict(json_dict["agent"]),
            scenario=Scenario.from_dict(json_dict["scenario"]),
            model=LanguageModel.from_dict(json_dict["model"]),
            iteration=json_dict["iteration"],
            answer=json_dict["answer"],
        )

    @property
    def agents(self):
        return [r.agent for r in self]


if __name__ == "__main__":
    import json

    print("Being imported")
    json_string = """
    {
    "survey": {
        "questions": [
            {
                "question_name": "how_feeling",
                "question_text": "How are you this {{ period }}?",
                "question_options": [
                    "Good",
                    "Great",
                    "OK",
                    "Bad"
                ],
                "type": "multiple_choice"
            }
        ],
        "name": null,
        "rule_collection": [
            {
                "current_q": 0,
                "expression": "True",
                "next_q": 1,
                "priority": -1,
                "question_name_to_index": {
                    "how_feeling": 0
                }
            }
        ]
    },
    "agent": {
        "traits": {
            "status": "Unhappy"
        },
        "verbose": false
    },
    "scenario": {
        "period": "morning"
    },
    "model": {
        "model": "gpt-3.5-turbo",
        "parameters": {
            "temperature": 0.5,
            "max_tokens": 1000,
            "top_p": 1,
            "frequency_penalty": 0,
            "presence_penalty": 0,
            "use_cache": true
        },
    },
    "iteration": 0,
    "answer": {
        "how_feeling": "Bad"
    }
    }
    """
    result = Result.from_dict(json.loads(json_string))

    # get combined dict working
    assert result.combined_dict["how_feeling"] == "Bad"
    assert result.get_value("answer", "how_feeling") == "Bad"

    # assert(result.get_value())



import textwrap
from pydantic import BaseModel, Field, field_validator
from typing import Type, Optional, Union
from edsl.questions.Question import Question
from edsl.questions.schemas import QuestionData, AnswerData
from edsl.questions.settings import Settings


class QuestionExtractDataTuples(Question):
    """Extracts a set of specified data types from a list of texts, storing information
    extracted from each text as a tuple, and then returns a list of those tuples."""

    question_type = "extract_data_tuples"

    @property
    def instructions(self):
        return textwrap.dedent(
            """\
        You are being asked to extract certain information from a set of texts.
        {{question_text}}:
        {% for d in data %}
        {{ loop.index0 }}: {{d}}
        {% endfor %} 
        From each text, extract the following information as a new comma-separated list
        with a place for each item in order: 
        {% for e in extract_data %}
        {{e}}
        {% endfor %} 
        If a text is missing any information, put an empty string in its place in the list,
        e.g., ["extracted info 1", "extracted info 2", "", "extracted info 4"]
        Return all of the lists of extracted information in a valid JSON formatted exactly like this: 
        {"answer": [<put your lists here separated by commas>], "comment": "<put explanation here>"}         
        """
        )

    def __repr__(self):
        return f"""{self.__class__.__name__}(question_text = "{self.question_text}", extract_data = "{self.extract_data}", question_name = "{self.question_name}")"""

    @classmethod
    def construct_question_data_model(cls) -> Type[BaseModel]:
        class LocalQuestionData(QuestionData):
            """Pydantic data model for QuestionExtractData"""

            question_text: Optional[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            extract_data: list[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            data: list[str] = Field(
                ..., min_length=1, max_length=Settings.MAX_QUESTION_LENGTH
            )
            allow_nonresponse: bool = None

        return LocalQuestionData

    def translate_answer_code_to_answer(self, answer, scenario=None):
        """There is no answer code."""
        return answer

    def construct_answer_data_model(self) -> Type[BaseModel]:
        class LocalAnswerDataModel(AnswerData):
            answer: list[list[Union[str, int, float]]] = Field(
                ..., min_length=0, max_length=Settings.MAX_ANSWER_LENGTH
            )

            @field_validator("answer")
            def check_answer(cls, value):
                if (
                    hasattr(self, "allow_nonresponse")
                    and self.allow_nonresponse == False
                    and (value == "" or value is None)
                ):
                    raise ValueError("You must provide a response.")
                return value

        return LocalAnswerDataModel

    def simulate_answer(self):
        "Simulates a valid answer for debugging purposes (what the validator expects)"
        raise NotImplementedError

    def form_elements(self):
        raise NotImplementedError

    # def __init__(self, question_text, extract_data, question_name=None):
    #     self.question_text = question_text
    #     self.extract_data = extract_data
    #     self.question_name = question_name

    #     d = {'question_text': question_text,
    #          'exptract_data': extract_data,
    #          'question_name': question_name}

    #     super().__init__(**d)


if __name__ == "__main__":
    from edsl import Agent, Scenario, print_dict_with_rich

    letters = [
        "Dear Ada, I am writing you a letter to tell you something...",
        "Dear Paul, I am writing you a letter to tell you something...",
        "Hi Penelope, I have some updates about my recent trip...",
    ]

    q = QuestionExtractDataTuples(
        question_text="The texts are letters",
        extract_data=[
            "the first initial of the person the letter is addressed to",
            "the name of the person the letter is addressed to",
            "the number of words in the letter",
        ],
        data=letters,
        question_name="letters_names",
    )
    scenarios = [Scenario({"letter": letter}) for letter in letters]
    print(q.get_prompt())
    results = q.by(*scenarios).run()
    print(results)

    # answer = {'answer': ["Ada", "Paul"]} #, 'comment': 'They are names that I found in the letters.'}
    # print(q.validate_answer(answer))
    # print(q.get_prompt())
